# gemato: OpenPGP signature support tests
# (c) 2017-2025 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import contextlib
import datetime
import io
import logging
import os
import shlex
import signal
import subprocess
import tempfile

from functools import cache
from pathlib import Path

import pytest

import gemato.cli
from gemato.compression import open_potentially_compressed_path
from gemato.exceptions import (
    ManifestUnsignedData,
    ManifestSyntaxError,
    OpenPGPNoImplementation,
    OpenPGPVerificationFailure,
    OpenPGPExpiredKeyFailure,
    OpenPGPRevokedKeyFailure,
    OpenPGPKeyImportError,
    OpenPGPKeyListingError,
    OpenPGPKeyRefreshError,
    OpenPGPRuntimeError,
    OpenPGPUntrustedSigFailure,
    ManifestInsecureHashes,
    )
from gemato.manifest import ManifestFile
from gemato.openpgp import (
    GNUPG,
    SystemGPGEnvironment,
    IsolatedGPGEnvironment,
    PGPyEnvironment,
    get_wkd_url,
    OpenPGPSignatureList,
    OpenPGPSignatureData,
    OpenPGPSignatureStatus,
    )
from gemato.recursiveloader import ManifestRecursiveLoader

from tests.test_recursiveloader import INSECURE_HASH_TESTS
from tests.testutil import HKPServer


data_dir = Path(__file__).parent / "data"


def break_sig(sig):
    """Return signature packet mangled to mismatch the signed key"""
    return sig[:-1] + b'\x55'


def dash_escape(data: str) -> str:
    """Add dash escapes to Manifest"""
    return "".join(
        [
            f"- {x}\n"
            if x.startswith(("TIMESTAMP", "MANIFEST", "DATA", "DIST"))
            else f"{x}\n"
            for x in data.splitlines()
        ]
    )


def modify_manifest(data: str) -> str:
    """Return the "modified" Manifest variation"""
    return "".join(
        [
            x.split(" 0 ", 1)[0] + " 32\n"
            if x.startswith("DATA")
            else f"{x}\n"
            for x in data.splitlines()
        ]
    )


def _(path: str) -> bytes:
    return data_dir.joinpath(path).read_bytes()


def F(path: str) -> str:
    return data_dir.joinpath(path).read_text().strip()


def T(path: str) -> str:
    return data_dir.joinpath(path).read_text()


# == first key (used for most of the tests) ==
PUBLIC_KEY = _("first-key/pub")
SECRET_KEY = _("first-key/secret")
# (subkey optionally used)
PUBLIC_SUBKEY = _("first-key/subkey")
# main UID
UID = _("first-key/uid")
# UID without email
UID_NOEMAIL = _("first-key/uid-noemail")
# UID that's not valid UTF-8
UID_NONUTF = _("first-key/uid-nonutf")

# UID signatures without expiration date
PUBLIC_KEY_SIG = _("first-key/uid-sig")
PUBLIC_KEY_NOEMAIL_SIG = _("first-key/uid-noemail-sig")
PUBLIC_KEY_NONUTF_SIG = _("first-key/uid-nonutf-sig")
# subkey signature without expiration date
PUBLIC_SUBKEY_SIG = _("first-key/subkey-sig")
# main UID signature that expired
EXPIRED_KEY_SIG = _("first-key/expired-sig")
# main UID revocation signature
REVOCATION_SIG = _("first-key/revocation-sig")
# main UID signature without expiration that's newer than expired-sig
UNEXPIRE_SIG = _("first-key/unexpire-sig")

# == other key (used in WKD tests, has the same UID as first key) ==
OTHER_PUBLIC_KEY = _("other-key/pub")
OTHER_PUBLIC_KEY_UID = _("other-key/uid")
OTHER_PUBLIC_KEY_SIG = _("other-key/uid-sig")

# == second key (using different UID) ==
SECOND_PUBLIC_KEY = _("second-key/pub")
SECOND_SECRET_KEY = _("second-key/secret")
SECOND_UID = _("second-key/uid")
SECOND_KEY_SIG = _("second-key/uid-sig")

# == combined variants of first-key ==
VALID_PUBLIC_KEY = PUBLIC_KEY + UID + PUBLIC_KEY_SIG
EXPIRED_PUBLIC_KEY = PUBLIC_KEY + UID + PUBLIC_KEY_SIG + EXPIRED_KEY_SIG
REVOKED_PUBLIC_KEY = PUBLIC_KEY + REVOCATION_SIG + UID + PUBLIC_KEY_SIG
# using the original signature to "unexpire" a key should fail
OLD_UNEXPIRE_PUBLIC_KEY = PUBLIC_KEY + UID + PUBLIC_KEY_SIG
UNEXPIRE_PUBLIC_KEY = PUBLIC_KEY + UID + UNEXPIRE_SIG

PRIVATE_KEY = SECRET_KEY + UID + PUBLIC_KEY_SIG

VALID_KEY_NOEMAIL = PUBLIC_KEY + UID_NOEMAIL + PUBLIC_KEY_NOEMAIL_SIG
VALID_KEY_NONUTF = PUBLIC_KEY + UID_NONUTF + PUBLIC_KEY_NONUTF_SIG

VALID_KEY_SUBKEY = (PUBLIC_KEY + UID + PUBLIC_KEY_SIG + PUBLIC_SUBKEY +
                    PUBLIC_SUBKEY_SIG)

# first-key with broken signature
FORGED_PUBLIC_KEY = PUBLIC_KEY + UID + break_sig(PUBLIC_KEY_SIG)
FORGED_SUBKEY = (PUBLIC_KEY + UID + PUBLIC_KEY_SIG + PUBLIC_SUBKEY +
                 break_sig(PUBLIC_SUBKEY_SIG))
FORGED_UNEXPIRE_KEY = (PUBLIC_KEY + UID + EXPIRED_KEY_SIG +
                       break_sig(UNEXPIRE_SIG))

# first-key without signatures
UNSIGNED_PUBLIC_KEY = PUBLIC_KEY + UID
UNSIGNED_SUBKEY = PUBLIC_KEY + UID + PUBLIC_KEY_SIG + PUBLIC_SUBKEY

# == key fingerprints ==
KEY_FINGERPRINT = F("first-key/fpr.txt")
SUBKEY_FINGERPRINT = F("first-key/sub-fpr.txt")
OTHER_KEY_FINGERPRINT = F("other-key/fpr.txt")
SECOND_KEY_FINGERPRINT = F("second-key/fpr.txt")

# == combined variants for other-key ==
OTHER_VALID_PUBLIC_KEY = (OTHER_PUBLIC_KEY + OTHER_PUBLIC_KEY_UID +
                          OTHER_PUBLIC_KEY_SIG)
COMBINED_PUBLIC_KEYS = OTHER_VALID_PUBLIC_KEY + VALID_PUBLIC_KEY

# == combined variants for second-key ==
SECOND_VALID_PUBLIC_KEY = SECOND_PUBLIC_KEY + SECOND_UID + SECOND_KEY_SIG
TWO_SIGNATURE_PUBLIC_KEYS = VALID_PUBLIC_KEY + SECOND_VALID_PUBLIC_KEY
TWO_KEYS_ONE_EXPIRED = EXPIRED_PUBLIC_KEY + SECOND_VALID_PUBLIC_KEY

# key with CRC error
MALFORMED_PUBLIC_KEY = _("malformed-key.txt")
# Manifest signed before first-key expired
SIGNED_MANIFEST = T("Manifest.asc")
# Manifest signed after first-key expired
POST_EXPIRATION_SIGNED_MANIFEST = T("Manifest.asc-post-expiration")
# valid Manifest with dash-escaped content
DASH_ESCAPED_SIGNED_MANIFEST = dash_escape(SIGNED_MANIFEST)
# Manifest with modified text (should fail)
MODIFIED_SIGNED_MANIFEST = modify_manifest(SIGNED_MANIFEST)
# Manifest with expired signature itself
EXPIRED_SIGNED_MANIFEST = T("Manifest.asc-expired")
# Manifest signed using the subkey
SUBKEY_SIGNED_MANIFEST = T("Manifest.asc-subkey-signed")
# combined signatures from first-key + second-key
TWO_SIGNATURE_MANIFEST = T("Manifest.asc-two-signatures")

# expected / base Manifest path (used for detached signature tests)
MANIFEST_PATH = data_dir / "Manifest"
# combined signatures from first-key + second-key
TWO_SIGNATURE_PATH = data_dir / "two-signatures.bin"


def strip_openpgp(text):
    lines = text.lstrip().splitlines()
    start = lines.index('')
    stop = lines.index('-----BEGIN PGP SIGNATURE-----')
    return '\n'.join(lines[start+1:stop-start+1]) + '\n'


@cache
def is_sequoia() -> bool:
    """Return True if GNUPG is sequoia-chameleon-gnupg"""
    try:
        out = subprocess.run([GNUPG, "--version"], capture_output=True)
    except OSError:
        return False
    return out.returncode == 0 and b"sequoia" in out.stdout


sequoia_xfail = pytest.mark.xfail(is_sequoia(),
                                  reason="FIXME on sequoia-chameleon-gnupg")


@pytest.mark.parametrize('manifest_var',
                         ["SIGNED_MANIFEST",
                          "DASH_ESCAPED_SIGNED_MANIFEST",
                          "SUBKEY_SIGNED_MANIFEST",
                          "MODIFIED_SIGNED_MANIFEST",
                          "EXPIRED_SIGNED_MANIFEST",
                          "TWO_SIGNATURE_MANIFEST",
                          "POST_EXPIRATION_SIGNED_MANIFEST",
                          ])
def test_noverify_goodish_manifest_load(manifest_var):
    """Test Manifest files that should succeed (OpenPGP disabled)"""
    m = ManifestFile()
    with io.StringIO(globals()[manifest_var]) as f:
        m.load(f, verify_openpgp=False)
    assert m.find_timestamp() is not None
    assert m.find_path_entry('myebuild-0.ebuild') is not None
    assert not m.openpgp_signed
    assert m.openpgp_signature is None


SIGNED_MANIFEST_JUNK_BEFORE = 'IGNORE test\n' + SIGNED_MANIFEST
SIGNED_MANIFEST_JUNK_AFTER = SIGNED_MANIFEST + 'IGNORE test\n'
SIGNED_MANIFEST_CUT_BEFORE_DATA = '\n'.join(
    SIGNED_MANIFEST.splitlines()[:3])
SIGNED_MANIFEST_CUT_BEFORE_SIGNATURE = '\n'.join(
    SIGNED_MANIFEST.splitlines()[:7])
SIGNED_MANIFEST_CUT_BEFORE_END = '\n'.join(
    SIGNED_MANIFEST.splitlines()[:15])


@pytest.mark.parametrize('manifest_var,expected',
                         [('SIGNED_MANIFEST_JUNK_BEFORE',
                           ManifestUnsignedData),
                          ('SIGNED_MANIFEST_JUNK_AFTER',
                           ManifestUnsignedData),
                          ('SIGNED_MANIFEST_CUT_BEFORE_DATA',
                           ManifestSyntaxError),
                          ('SIGNED_MANIFEST_CUT_BEFORE_SIGNATURE',
                           ManifestSyntaxError),
                          ('SIGNED_MANIFEST_CUT_BEFORE_END',
                           ManifestSyntaxError),
                          ])
def test_noverify_bad_manifest_load(manifest_var, expected):
    """Test Manifest files that should fail"""
    m = ManifestFile()
    with io.StringIO(globals()[manifest_var]) as f:
        with pytest.raises(expected):
            m.load(f, verify_openpgp=False)


@pytest.mark.parametrize('write_back', [False, True])
def test_noverify_recursive_manifest_loader(tmp_path, write_back):
    """Test reading signed Manifest"""
    with open(tmp_path / 'Manifest', 'w') as f:
        f.write(MODIFIED_SIGNED_MANIFEST)

    m = ManifestRecursiveLoader(tmp_path / 'Manifest',
                                verify_openpgp=False)
    assert not m.openpgp_signed
    assert m.openpgp_signature is None

    if write_back:
        m.save_manifest('Manifest')
        with open(tmp_path / 'Manifest') as f:
            assert f.read() == strip_openpgp(MODIFIED_SIGNED_MANIFEST)


def test_noverify_load_cli(tmp_path):
    """Test reading signed Manifest via CLI"""
    with open(tmp_path / 'Manifest', 'w') as f:
        f.write(MODIFIED_SIGNED_MANIFEST)
    os.mkdir(tmp_path / 'eclass')
    with open(tmp_path / 'eclass' / 'Manifest', 'w'):
        pass
    with open(tmp_path / 'myebuild-0.ebuild', 'wb') as f:
        f.write(b'12345678901234567890123456789012')
    with open(tmp_path / 'metadata.xml', 'wb'):
        pass

    assert 0 == gemato.cli.main(['gemato', 'verify',
                                 '--no-openpgp-verify', str(tmp_path)])


class MockedSystemGPGEnvironment(SystemGPGEnvironment):
    """System environment variant mocked to use isolated GNUPGHOME"""
    def __init__(self, *args, **kwargs):
        self._tmpdir = tempfile.TemporaryDirectory()
        self._home = self._tmpdir.name
        os.environ['GNUPGHOME'] = self._tmpdir.name
        super().__init__(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_cb):
        self.close()

    def close(self):
        if self._tmpdir is not None:
            IsolatedGPGEnvironment.close(self)
            # we need to recreate it to make cleanup() happy
            os.mkdir(self._tmpdir.name)
            self._tmpdir.cleanup()
            self._tmpdir = None
            os.environ.pop('GNUPGHOME', None)

    def import_key(self, keyfile, trust=True):
        IsolatedGPGEnvironment.import_key(self, keyfile, trust=trust)


@pytest.fixture(params=[IsolatedGPGEnvironment,
                        MockedSystemGPGEnvironment,
                        PGPyEnvironment,
                        ])
def openpgp_env(request):
    """OpenPGP environment fixture"""
    try:
        env = request.param()
    except OpenPGPNoImplementation as e:
        pytest.skip(str(e))
    yield env
    env.close()


@pytest.fixture(params=[IsolatedGPGEnvironment,
                        ])
def openpgp_env_with_refresh(request):
    """OpenPGP environments that support refreshing keys"""
    env = request.param()
    yield env
    env.close()


MANIFEST_VARIANTS = [
    # manifest, key, expected fpr/exception
    # == good manifests ==
    ('SIGNED_MANIFEST', 'VALID_PUBLIC_KEY', None),
    ('SIGNED_MANIFEST', 'VALID_KEY_NOEMAIL', None),
    ('SIGNED_MANIFEST', 'VALID_KEY_NONUTF', None),
    ('SIGNED_MANIFEST', 'COMBINED_PUBLIC_KEYS', None),
    ('DASH_ESCAPED_SIGNED_MANIFEST', 'VALID_PUBLIC_KEY', None),
    ('SUBKEY_SIGNED_MANIFEST', 'VALID_KEY_SUBKEY', None),
    ("POST_EXPIRATION_SIGNED_MANIFEST", "VALID_PUBLIC_KEY", None),
    # == Manifest signed before the key expired ==
    ("SIGNED_MANIFEST", "EXPIRED_PUBLIC_KEY", None),
    # == Manifest with two signatures ==
    ("TWO_SIGNATURE_MANIFEST", "TWO_SIGNATURE_PUBLIC_KEYS", None),
    ("TWO_SIGNATURE_MANIFEST", "VALID_PUBLIC_KEY", OpenPGPVerificationFailure),
    ("TWO_SIGNATURE_MANIFEST", "SECOND_VALID_PUBLIC_KEY",
     OpenPGPVerificationFailure),
    ("TWO_SIGNATURE_MANIFEST", "TWO_KEYS_ONE_EXPIRED",
     OpenPGPExpiredKeyFailure),
    # == using private key ==
    ('SIGNED_MANIFEST', 'PRIVATE_KEY', None),
    # == bad manifests ==
    ('MODIFIED_SIGNED_MANIFEST', 'VALID_PUBLIC_KEY',
     OpenPGPVerificationFailure),
    ('EXPIRED_SIGNED_MANIFEST', 'VALID_PUBLIC_KEY',
     OpenPGPVerificationFailure),
    # == bad keys ==
    ('SIGNED_MANIFEST', None,
     OpenPGPVerificationFailure),
    ("POST_EXPIRATION_SIGNED_MANIFEST", "EXPIRED_PUBLIC_KEY",
     OpenPGPExpiredKeyFailure),
    ('SIGNED_MANIFEST', 'REVOKED_PUBLIC_KEY',
     OpenPGPRevokedKeyFailure),
    ('SIGNED_MANIFEST', 'OTHER_VALID_PUBLIC_KEY',
     OpenPGPVerificationFailure),
    ('SIGNED_MANIFEST', 'UNSIGNED_PUBLIC_KEY',
     OpenPGPKeyImportError),
    ('SIGNED_MANIFEST', 'FORGED_PUBLIC_KEY',
     OpenPGPKeyImportError),
    ('SUBKEY_SIGNED_MANIFEST', 'UNSIGNED_SUBKEY',
     OpenPGPVerificationFailure),
    ('SUBKEY_SIGNED_MANIFEST', 'FORGED_SUBKEY',
     OpenPGPVerificationFailure),
]


def assert_signature(sig: OpenPGPSignatureList,
                     manifest_var: str,
                     expect_both: bool = True,
                     ) -> None:
    """Make assertions about the signature"""
    if manifest_var == "TWO_SIGNATURE_MANIFEST":
        no_key_sig = OpenPGPSignatureData(
            sig_status=OpenPGPSignatureStatus.NO_PUBLIC_KEY)
        assert sorted(sig) == sorted([
            OpenPGPSignatureData(
                fingerprint=KEY_FINGERPRINT,
                timestamp=datetime.datetime(2024, 1, 1),
                primary_key_fingerprint=KEY_FINGERPRINT,
                sig_status=OpenPGPSignatureStatus.GOOD,
                trusted_sig=True,
                valid_sig=True,
                ),
            OpenPGPSignatureData(
                fingerprint=SECOND_KEY_FINGERPRINT,
                timestamp=datetime.datetime(2020, 1, 1),
                primary_key_fingerprint=SECOND_KEY_FINGERPRINT,
                sig_status=OpenPGPSignatureStatus.GOOD,
                trusted_sig=True,
                valid_sig=True,
                ) if expect_both else no_key_sig,
        ])
    elif manifest_var == 'SUBKEY_SIGNED_MANIFEST':
        assert len(sig) == 1
        assert sig.fingerprint == SUBKEY_FINGERPRINT
        assert sig.timestamp == datetime.datetime(2020, 1, 1)
        assert sig.expire_timestamp is None
        assert sig.primary_key_fingerprint == KEY_FINGERPRINT
    elif manifest_var == "POST_EXPIRATION_SIGNED_MANIFEST":
        assert len(sig) == 1
        assert sig.fingerprint == KEY_FINGERPRINT
        assert sig.timestamp == datetime.datetime(2024, 1, 1)
        assert sig.expire_timestamp is None
        assert sig.primary_key_fingerprint == KEY_FINGERPRINT
    else:
        assert len(sig) == 1
        assert sig.fingerprint == KEY_FINGERPRINT
        assert sig.timestamp == datetime.datetime(2020, 1, 1)
        assert sig.expire_timestamp is None
        assert sig.primary_key_fingerprint == KEY_FINGERPRINT


@pytest.mark.parametrize('manifest_var,key_var,expected',
                         MANIFEST_VARIANTS)
def test_verify_manifest(openpgp_env, manifest_var, key_var, expected):
    """Test direct Manifest data verification"""
    if (isinstance(openpgp_env, PGPyEnvironment) and
            manifest_var == 'DASH_ESCAPED_SIGNED_MANIFEST'):
        pytest.xfail('dash escaping is known-broken in pgpy')

    try:
        with io.StringIO(globals()[manifest_var]) as f:
            if expected is None:
                if key_var is not None:
                    with io.BytesIO(globals()[key_var]) as kf:
                        openpgp_env.import_key(kf)

                sig = openpgp_env.verify_file(f)
                assert_signature(sig, manifest_var)
            else:
                with pytest.raises(expected):
                    if key_var is not None:
                        with io.BytesIO(globals()[key_var]) as kf:
                            openpgp_env.import_key(kf)

                    openpgp_env.verify_file(f)
    except OpenPGPNoImplementation as e:
        pytest.skip(str(e))


def test_verify_one_out_of_two():
    try:
        with MockedSystemGPGEnvironment() as openpgp_env:
            with io.BytesIO(VALID_PUBLIC_KEY) as f:
                openpgp_env.import_key(f)

            with io.StringIO(TWO_SIGNATURE_MANIFEST) as f:
                sig = openpgp_env.verify_file(f, require_all_good=False)

            assert_signature(sig, "TWO_SIGNATURE_MANIFEST", expect_both=False)
    except OpenPGPNoImplementation as e:
        pytest.skip(str(e))


def test_verify_untrusted_key():
    try:
        with MockedSystemGPGEnvironment() as openpgp_env:
            with io.BytesIO(VALID_PUBLIC_KEY) as f:
                openpgp_env.import_key(f, trust=False)

            with io.StringIO(SIGNED_MANIFEST) as f:
                with pytest.raises(OpenPGPUntrustedSigFailure):
                    openpgp_env.verify_file(f)
    except OpenPGPNoImplementation as e:
        pytest.skip(str(e))


@pytest.mark.parametrize('manifest_var,key_var,expected',
                         MANIFEST_VARIANTS)
def test_manifest_load(openpgp_env, manifest_var, key_var, expected):
    """Test Manifest verification via ManifestFile.load()"""
    if (isinstance(openpgp_env, PGPyEnvironment) and
            manifest_var == 'DASH_ESCAPED_SIGNED_MANIFEST'):
        pytest.xfail('dash escaping is known-broken in pgpy')

    try:
        key_loaded = False
        m = ManifestFile()
        with io.StringIO(globals()[manifest_var]) as f:
            if expected is None:
                if key_var is not None:
                    with io.BytesIO(globals()[key_var]) as kf:
                        openpgp_env.import_key(kf)

                key_loaded = True
                m.load(f, openpgp_env=openpgp_env)
                assert m.openpgp_signed
                assert_signature(m.openpgp_signature, manifest_var)
            else:
                with pytest.raises(expected):
                    if key_var is not None:
                        with io.BytesIO(globals()[key_var]) as kf:
                            openpgp_env.import_key(kf)

                    key_loaded = True
                    m.load(f, openpgp_env=openpgp_env)
                assert not m.openpgp_signed
                assert m.openpgp_signature is None

        if key_loaded:
            # Manifest entries should be loaded even if verification failed
            assert m.find_timestamp() is not None
            assert m.find_path_entry('myebuild-0.ebuild') is not None
    except OpenPGPNoImplementation as e:
        pytest.skip(str(e))


@pytest.mark.parametrize('filename', ['Manifest', 'Manifest.gz'])
@pytest.mark.parametrize('manifest_var,key_var,expected',
                         MANIFEST_VARIANTS)
def test_recursive_manifest_loader(tmp_path, openpgp_env, filename,
                                   manifest_var, key_var, expected):
    """Test Manifest verification via ManifestRecursiveLoader"""
    if (isinstance(openpgp_env, PGPyEnvironment) and
            manifest_var == 'DASH_ESCAPED_SIGNED_MANIFEST'):
        pytest.xfail('dash escaping is known-broken in pgpy')

    try:
        with open_potentially_compressed_path(tmp_path / filename, 'w') as cf:
            cf.write(globals()[manifest_var])

        if expected is None:
            if key_var is not None:
                with io.BytesIO(globals()[key_var]) as f:
                    openpgp_env.import_key(f)

            m = ManifestRecursiveLoader(tmp_path / filename,
                                        verify_openpgp=True,
                                        openpgp_env=openpgp_env)
            assert m.openpgp_signed
            assert_signature(m.openpgp_signature, manifest_var)
        else:
            with pytest.raises(expected):
                if key_var is not None:
                    with io.BytesIO(globals()[key_var]) as f:
                        openpgp_env.import_key(f)

                ManifestRecursiveLoader(tmp_path / filename,
                                        verify_openpgp=True,
                                        openpgp_env=openpgp_env)
    except OpenPGPNoImplementation as e:
        pytest.skip(str(e))


@pytest.fixture
def base_tree(tmp_path):
    os.mkdir(tmp_path / 'eclass')
    with open(tmp_path / 'eclass' / 'Manifest', 'w'):
        pass
    with open(tmp_path / 'myebuild-0.ebuild', 'wb'):
        pass
    with open(tmp_path / 'metadata.xml', 'wb'):
        pass
    return tmp_path


@pytest.mark.parametrize('manifest_var,key_var,expected',
                         [(m, k, e) for m, k, e in MANIFEST_VARIANTS
                          if k is not None])
def test_cli(base_tree, caplog, manifest_var, key_var, expected):
    """Test Manifest verification via CLI"""
    with open(base_tree / '.key.bin', 'wb') as f:
        f.write(globals()[key_var])
    with open(base_tree / 'Manifest', 'w') as f:
        f.write(globals()[manifest_var])
    if manifest_var == 'MODIFIED_SIGNED_MANIFEST':
        with open(base_tree / 'myebuild-0.ebuild', 'wb') as f:
            f.write(b'12345678901234567890123456789012')

    retval = gemato.cli.main(['gemato', 'verify',
                              '--openpgp-key',
                              str(base_tree / '.key.bin'),
                              '--no-refresh-keys',
                              '--require-signed-manifest',
                              # we verify this option separately
                              # and our test data currently sucks
                              '--no-require-secure-hashes',
                              str(base_tree)])
    if str(OpenPGPNoImplementation('install gpg')) in caplog.text:
        pytest.skip('OpenPGP implementation missing')

    eexit = 0 if expected is None else 1
    assert retval == eexit
    if expected is not None:
        assert str(expected('')) in caplog.text


EMPTY_DATA = b''


@pytest.mark.parametrize(
    'key_var,success',
    [('VALID_PUBLIC_KEY', True),
     ('VALID_KEY_NOEMAIL', True),
     ('VALID_KEY_NONUTF', True),
     ('MALFORMED_PUBLIC_KEY', False),
     ('EMPTY_DATA', False),
     ('FORGED_PUBLIC_KEY', False),
     ('UNSIGNED_PUBLIC_KEY', False),
     ])
def test_env_import_key(openpgp_env, key_var, success):
    """Test importing valid and invalid keys"""
    try:
        if success:
            openpgp_env.import_key(io.BytesIO(globals()[key_var]))
        else:
            with pytest.raises(OpenPGPKeyImportError):
                openpgp_env.import_key(io.BytesIO(globals()[key_var]))
    except OpenPGPNoImplementation as e:
        pytest.skip(str(e))


def test_env_double_close():
    """Test that env can be closed multiple times"""
    with IsolatedGPGEnvironment() as env:
        env.close()


def test_env_home_after_close():
    """Test that .home can not be referenced after closing"""
    with IsolatedGPGEnvironment() as env:
        env.close()
        with pytest.raises(AssertionError):
            env.home


@pytest.fixture(params=[IsolatedGPGEnvironment,
                        MockedSystemGPGEnvironment,
                        ])
def privkey_env(request):
    """Environment with private key loaded"""
    try:
        env = request.param()
        env.import_key(io.BytesIO(PRIVATE_KEY))
    except OpenPGPNoImplementation as e:
        pytest.skip(str(e))
    yield env
    env.close()


TEST_STRING = 'The quick brown fox jumps over the lazy dog'


@sequoia_xfail
@pytest.mark.parametrize('keyid', [None, KEY_FINGERPRINT])
def test_sign_data(privkey_env, keyid):
    """Test signing data"""
    with io.StringIO(TEST_STRING) as f:
        with io.StringIO() as wf:
            privkey_env.clear_sign_file(f, wf, keyid=keyid)
            wf.seek(0)
            privkey_env.verify_file(wf)


@pytest.mark.parametrize('keyid', [None, KEY_FINGERPRINT])
@pytest.mark.parametrize('sign', [None, False, True])
def test_dump_signed_manifest(privkey_env, keyid, sign):
    """Test dumping a signed Manifest"""
    if sign is not False and is_sequoia():
        pytest.xfail("FIXME: sequoia-chameleon-gnupg fails to sign")

    m = ManifestFile()
    verify = True if sign is None else False
    with io.StringIO(SIGNED_MANIFEST) as f:
        m.load(f, verify_openpgp=verify, openpgp_env=privkey_env)
    assert m.openpgp_signed == verify

    with io.StringIO() as f:
        m.dump(f, openpgp_keyid=keyid, openpgp_env=privkey_env,
               sign_openpgp=sign)
        f.seek(0)
        m.load(f, openpgp_env=privkey_env)
    if sign is not False:
        assert m.openpgp_signed
        assert m.openpgp_signature is not None
    else:
        assert not m.openpgp_signed
        assert m.openpgp_signature is None


@sequoia_xfail
@pytest.mark.parametrize('filename', ['Manifest', 'Manifest.gz'])
@pytest.mark.parametrize('sign', [None, True])
def test_recursive_manifest_loader_save_manifest(tmp_path, privkey_env,
                                                 filename, sign):
    """Test signing Manifests via ManifestRecursiveLoader"""
    with open_potentially_compressed_path(tmp_path / filename, 'w') as cf:
        cf.write(SIGNED_MANIFEST)

    verify = not sign
    m = ManifestRecursiveLoader(tmp_path / filename,
                                verify_openpgp=verify,
                                sign_openpgp=sign,
                                openpgp_env=privkey_env)
    assert m.openpgp_signed == verify

    m.save_manifest(filename)
    m2 = ManifestFile()
    with open_potentially_compressed_path(tmp_path / filename, 'r') as cf:
        m2.load(cf, openpgp_env=privkey_env)
    assert m2.openpgp_signed
    assert m2.openpgp_signature is not None


def test_recursive_manifest_loader_save_submanifest(tmp_path, privkey_env):
    """Test that sub-Manifests are not signed"""
    with open(tmp_path / 'Manifest', 'w') as f:
        f.write(SIGNED_MANIFEST)
    os.mkdir(tmp_path / 'eclass')
    with open(tmp_path / 'eclass' / 'Manifest', 'w'):
        pass

    m = ManifestRecursiveLoader(tmp_path / 'Manifest',
                                verify_openpgp=False,
                                sign_openpgp=True,
                                openpgp_env=privkey_env)
    assert not m.openpgp_signed
    assert m.openpgp_signature is None

    m.load_manifest('eclass/Manifest')
    m.save_manifest('eclass/Manifest')

    m2 = ManifestFile()
    with open(tmp_path / 'eclass' / 'Manifest') as f:
        m2.load(f, openpgp_env=privkey_env)
    assert not m2.openpgp_signed
    assert m2.openpgp_signature is None


@pytest.mark.parametrize(
    'key_var,expected',
    [('VALID_PUBLIC_KEY',
      {KEY_FINGERPRINT: [b"gemato test key <gemato@example.com>"]}),
     ('OTHER_VALID_PUBLIC_KEY',
      {OTHER_KEY_FINGERPRINT: [b"gemato test key <gemato@example.com>"]}),
     ('VALID_KEY_SUBKEY',
      {KEY_FINGERPRINT: [b"gemato test key <gemato@example.com>"]}),
     ('VALID_KEY_NOEMAIL',
      {KEY_FINGERPRINT: [b"gemato test key"]}),
     ('VALID_KEY_NONUTF',
      {KEY_FINGERPRINT: [
           b"gemat\\xf6 test key <gemato@example.com>" if is_sequoia()
           else b"gemat\xf6 test key <gemato@example.com>"
       ]}),
     ])
def test_list_keys(openpgp_env, key_var, expected):
    try:
        openpgp_env.import_key(io.BytesIO(globals()[key_var]))
    except OpenPGPNoImplementation as e:
        pytest.skip(str(e))
    assert openpgp_env.list_keys() == expected
    assert openpgp_env.list_keys(list(expected.keys())) == expected


def test_list_keys_empty(openpgp_env):
    try:
        assert openpgp_env.list_keys() == {}
    except OpenPGPNoImplementation as e:
        pytest.skip(str(e))


def test_list_keys_error(openpgp_env):
    try:
        with pytest.raises(OpenPGPKeyListingError):
            openpgp_env.list_keys(["zzzzz"])
    except OpenPGPNoImplementation as e:
        pytest.skip(str(e))


@pytest.fixture(scope='module')
def global_hkp_server():
    """A fixture that starts a single HKP server instance for tests"""
    server = HKPServer()
    server.start()
    yield server
    server.stop()


@pytest.fixture
def hkp_server(global_hkp_server):
    """A fixture that resets the global HKP server with empty keys"""
    global_hkp_server.keys.clear()
    yield global_hkp_server


REFRESH_VARIANTS = [
    # manifest, key, server key fpr, server key, expected exception
    ('SIGNED_MANIFEST', 'VALID_PUBLIC_KEY', KEY_FINGERPRINT,
     'VALID_PUBLIC_KEY', None),
    ('SIGNED_MANIFEST', 'VALID_KEY_NONUTF', KEY_FINGERPRINT,
     'VALID_PUBLIC_KEY', None),
    ('SIGNED_MANIFEST', 'VALID_PUBLIC_KEY', KEY_FINGERPRINT,
     'REVOKED_PUBLIC_KEY', OpenPGPRevokedKeyFailure),
    # test fetching subkey for primary key
    ('SUBKEY_SIGNED_MANIFEST', 'VALID_PUBLIC_KEY', KEY_FINGERPRINT,
     'VALID_KEY_SUBKEY', None),
    # refresh should fail if key is not on server
    ('SIGNED_MANIFEST', 'VALID_PUBLIC_KEY', None, None,
     OpenPGPKeyRefreshError),
    # unrevocation should not be possible
    ('SIGNED_MANIFEST', 'REVOKED_PUBLIC_KEY', KEY_FINGERPRINT,
     'VALID_PUBLIC_KEY', OpenPGPRevokedKeyFailure),
    # unexpiration should be possible
    ('SIGNED_MANIFEST', 'EXPIRED_PUBLIC_KEY', KEY_FINGERPRINT,
     'UNEXPIRE_PUBLIC_KEY', None),
    # ...but only with a new signature
    ("POST_EXPIRATION_SIGNED_MANIFEST", "EXPIRED_PUBLIC_KEY", KEY_FINGERPRINT,
     "OLD_UNEXPIRE_PUBLIC_KEY", OpenPGPExpiredKeyFailure),
    # make sure server can't malicously inject or replace key
    ('SIGNED_MANIFEST', 'OTHER_VALID_PUBLIC_KEY', OTHER_KEY_FINGERPRINT,
     'VALID_PUBLIC_KEY', OpenPGPKeyRefreshError),
    ('SIGNED_MANIFEST', 'OTHER_VALID_PUBLIC_KEY', OTHER_KEY_FINGERPRINT,
     'COMBINED_PUBLIC_KEYS', OpenPGPRuntimeError),
    # test that forged keys are rejected
    ("POST_EXPIRATION_SIGNED_MANIFEST", "EXPIRED_PUBLIC_KEY", KEY_FINGERPRINT,
     "FORGED_UNEXPIRE_KEY", OpenPGPExpiredKeyFailure),
    ('SUBKEY_SIGNED_MANIFEST', 'VALID_PUBLIC_KEY', KEY_FINGERPRINT,
     'FORGED_SUBKEY', OpenPGPVerificationFailure),
    ('SUBKEY_SIGNED_MANIFEST', 'VALID_PUBLIC_KEY', KEY_FINGERPRINT,
     'UNSIGNED_SUBKEY', OpenPGPVerificationFailure),
]


@pytest.mark.parametrize(
    'manifest_var,key_var,server_key_fpr,server_key_var,expected',
    REFRESH_VARIANTS +
    [('SIGNED_MANIFEST', 'VALID_KEY_NOEMAIL', KEY_FINGERPRINT,
      'VALID_PUBLIC_KEY', None),
     ])
def test_refresh_hkp(openpgp_env_with_refresh, hkp_server, manifest_var,
                     key_var, server_key_fpr, server_key_var, expected):
    """Test refreshing against a HKP keyserver"""
    try:
        if key_var is not None:
            with io.BytesIO(globals()[key_var]) as f:
                openpgp_env_with_refresh.import_key(f)

        if server_key_var is not None:
            hkp_server.keys[server_key_fpr] = globals()[server_key_var]

        if expected is None:
            openpgp_env_with_refresh.refresh_keys(
                allow_wkd=False, keyserver=hkp_server.addr)
            with io.StringIO(globals()[manifest_var]) as f:
                openpgp_env_with_refresh.verify_file(f)
        else:
            with pytest.raises(expected):
                openpgp_env_with_refresh.refresh_keys(
                    allow_wkd=False, keyserver=hkp_server.addr)
                with io.StringIO(globals()[manifest_var]) as f:
                    openpgp_env_with_refresh.verify_file(f)
    except OpenPGPNoImplementation as e:
        pytest.skip(str(e))


@pytest.mark.parametrize(
    'manifest_var,key_var,server_key_fpr,server_key_var,expected,'
    'expect_hit',
    [args + (True,) for args in REFRESH_VARIANTS] +
    [('SIGNED_MANIFEST', 'VALID_KEY_NOEMAIL', KEY_FINGERPRINT,
      'VALID_PUBLIC_KEY', OpenPGPKeyRefreshError, False),
     ])
def test_refresh_wkd(openpgp_env_with_refresh,
                     manifest_var,
                     key_var,
                     server_key_fpr,
                     server_key_var,
                     expected,
                     expect_hit):
    """Test refreshing against WKD"""
    with pytest.importorskip('responses').RequestsMock(
            assert_all_requests_are_fired=expect_hit) as responses:
        try:
            if key_var is not None:
                with io.BytesIO(globals()[key_var]) as f:
                    openpgp_env_with_refresh.import_key(f)

            if server_key_var is not None:
                responses.add(
                    responses.GET,
                    'https://example.com/.well-known/openpgpkey/hu/'
                    '5x66h616iaskmnadrm86ndo6xnxbxjxb?l=gemato',
                    body=globals()[server_key_var],
                    content_type='application/pgp-keys')
            else:
                responses.add(
                    responses.GET,
                    'https://example.com/.well-known/openpgpkey/hu/'
                    '5x66h616iaskmnadrm86ndo6xnxbxjxb?l=gemato',
                    status=404)

            if expected is None:
                openpgp_env_with_refresh.refresh_keys(
                    allow_wkd=True, keyserver='hkps://block.invalid/')
                with io.StringIO(globals()[manifest_var]) as f:
                    openpgp_env_with_refresh.verify_file(f)
            else:
                with pytest.raises(expected):
                    openpgp_env_with_refresh.refresh_keys(
                        allow_wkd=True, keyserver='hkps://block.invalid/')
                    with io.StringIO(globals()[manifest_var]) as f:
                        openpgp_env_with_refresh.verify_file(f)
        except OpenPGPNoImplementation as e:
            pytest.skip(str(e))


@pytest.mark.parametrize('status', [401, 404, 500, ConnectionError])
def test_refresh_wkd_fallback_to_hkp(openpgp_env_with_refresh,
                                     hkp_server, caplog, status):
    """Test whether WKD refresh failure falls back to HKP"""
    with pytest.importorskip('responses').RequestsMock() as responses:
        try:
            with io.BytesIO(VALID_PUBLIC_KEY) as f:
                openpgp_env_with_refresh.import_key(f)
            hkp_server.keys[KEY_FINGERPRINT] = REVOKED_PUBLIC_KEY
            if status is not ConnectionError:
                responses.add(
                    responses.GET,
                    'https://example.com/.well-known/openpgpkey/hu/'
                    '5x66h616iaskmnadrm86ndo6xnxbxjxb?l=gemato',
                    status=status)

            caplog.set_level(logging.DEBUG)
            openpgp_env_with_refresh.refresh_keys(
                allow_wkd=True, keyserver=hkp_server.addr)
            assert 'failing due to failed request' in caplog.text

            with pytest.raises(OpenPGPRevokedKeyFailure):
                with io.StringIO(SIGNED_MANIFEST) as f:
                    openpgp_env_with_refresh.verify_file(f)
        except OpenPGPNoImplementation as e:
            pytest.skip(str(e))


@pytest.mark.parametrize(
    'email,expected',
    [('gemato@example.com',
      'https://example.com/.well-known/openpgpkey/hu/'
      '5x66h616iaskmnadrm86ndo6xnxbxjxb?l=gemato'),
     ('Joe.Doe@Example.ORG',
      'https://example.org/.well-known/openpgpkey/hu/'
      'iy9q119eutrkn8s1mk4r39qejnbu3n5q?l=Joe.Doe'),
     ])
def test_get_wkd_url(email, expected):
    assert get_wkd_url(email) == expected


def signal_desc(sig):
    if hasattr(signal, 'strsignal'):
        return signal.strsignal(sig)
    else:
        return sig


@pytest.mark.parametrize(
    'command,expected,match',
    [('true', 0, None),
     ('false', 1, None),
     ('{gpg} --verify {tmp_path}/Manifest', 0, None),
     ('{gpg} --verify {tmp_path}/Manifest.subkey', 2, None),
     ('sh -c "kill $$"', -signal.SIGTERM,
      f'Child process terminated due to signal: '
      f'{signal_desc(signal.SIGTERM)}'),
     ('sh -c "kill -USR1 $$"', -signal.SIGUSR1,
      f'Child process terminated due to signal: '
      f'{signal_desc(signal.SIGUSR1)}'),
     ])
def test_cli_gpg_wrap(tmp_path, caplog, command, expected, match):
    with open(tmp_path / '.key.bin', 'wb') as f:
        f.write(VALID_PUBLIC_KEY)
    with open(tmp_path / 'Manifest', 'w') as f:
        f.write(SIGNED_MANIFEST)
    with open(tmp_path / 'Manifest.subkey', 'w') as f:
        f.write(SUBKEY_SIGNED_MANIFEST)

    command = [x.replace('{tmp_path}', str(tmp_path)).replace('{gpg}', GNUPG)
               for x in shlex.split(command)]
    retval = gemato.cli.main(['gemato', 'gpg-wrap',
                              '--openpgp-key',
                              str(tmp_path / '.key.bin'),
                              '--no-refresh-keys',
                              '--'] + command)
    if str(OpenPGPNoImplementation('install gpg')) in caplog.text:
        pytest.skip('OpenPGP implementation missing')

    assert retval == expected
    if match is not None:
        assert match in caplog.text


@pytest.mark.parametrize("hashes_arg,insecure", INSECURE_HASH_TESTS)
@pytest.mark.parametrize(
    "sign,require_secure",
    [(None, None),
     (False, None),
     (True, None),
     (None, False),
     (True, False),
     ])
def test_recursive_manifest_loader_require_secure(tmp_path, privkey_env,
                                                  hashes_arg, insecure,
                                                  sign, require_secure):
    with open(tmp_path / "Manifest", "w") as f:
        f.write(SIGNED_MANIFEST)

    ctx = (pytest.raises(ManifestInsecureHashes)
           if insecure is not None and sign is not False
           and require_secure is not False
           else contextlib.nullcontext())
    with ctx:
        m = ManifestRecursiveLoader(tmp_path / "Manifest",
                                    hashes=hashes_arg.split(),
                                    require_secure_hashes=require_secure,
                                    verify_openpgp=not sign,
                                    sign_openpgp=sign,
                                    openpgp_env=privkey_env)
        if not sign:
            assert m.openpgp_signed


@pytest.mark.parametrize("hashes_arg,insecure", INSECURE_HASH_TESTS)
@pytest.mark.parametrize(
    "sign,require_secure",
    [("", ""),
     ("--no-sign", ""),
     ("--sign", ""),
     ("", "--no-require-secure-hashes"),
     ("--sign", "--no-require-secure-hashes"),
     ])
def test_update_require_secure_cli(base_tree, caplog, hashes_arg,
                                   insecure, sign, require_secure):
    expected = (1 if insecure is not None and sign != "--no-sign"
                and require_secure != "--no-require-secure-hashes"
                else 0)
    if expected == 0 and is_sequoia():
        pytest.xfail("FIXME: sequoia-chameleon-gnupg fails to sign")

    with open(base_tree / ".key.bin", "wb") as keyf:
        keyf.write(PRIVATE_KEY)
    with open(base_tree / "Manifest", "w") as f:
        f.write(SIGNED_MANIFEST)

    retval = gemato.cli.main(["gemato", "update",
                              "-K", str(base_tree / ".key.bin"),
                              "--hashes", hashes_arg,
                              str(base_tree)]
                             + f"{sign} {require_secure}".split())
    if str(OpenPGPNoImplementation('install gpg')) in caplog.text:
        pytest.skip('OpenPGP implementation missing')

    assert retval == expected
    if expected == 1:
        assert str(ManifestInsecureHashes(insecure)) in caplog.text


@pytest.mark.parametrize(
    "require_secure", ["", "--no-require-secure-hashes"])
def test_verify_require_secure_cli(base_tree, caplog, require_secure):
    with open(base_tree / ".key.bin", "wb") as keyf:
        keyf.write(VALID_PUBLIC_KEY)
    with open(base_tree / "Manifest", "w") as f:
        f.write(SIGNED_MANIFEST)

    retval = gemato.cli.main(["gemato", "verify",
                              "--no-refresh-keys",
                              "--require-signed-manifest",
                              "-K", str(base_tree / ".key.bin"),
                              str(base_tree)]
                             + require_secure.split())
    if str(OpenPGPNoImplementation('install gpg')) in caplog.text:
        pytest.skip('OpenPGP implementation missing')

    expected = (1 if require_secure != "--no-require-secure-hashes"
                else 0)
    assert retval == expected
    if expected == 1:
        assert str(ManifestInsecureHashes(["MD5"])) in caplog.text


@pytest.mark.parametrize(
    "key_var,two_sigs",
    [("TWO_SIGNATURE_PUBLIC_KEYS", True),
     ("VALID_PUBLIC_KEY", False),
     ])
def test_verify_detached(tmp_path, key_var, two_sigs):
    try:
        with MockedSystemGPGEnvironment() as openpgp_env:
            with io.BytesIO(globals()[key_var]) as f:
                openpgp_env.import_key(f)

            with open(MANIFEST_PATH, "rb") as f:
                sig = openpgp_env.verify_detached(
                    TWO_SIGNATURE_PATH, f,
                    require_all_good=two_sigs)

            assert_signature(sig, "TWO_SIGNATURE_MANIFEST",
                             expect_both=two_sigs)
    except OpenPGPNoImplementation as e:
        pytest.skip(str(e))



# gemato: File verification routines
# (c) 2017-2022 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import contextlib
import errno
import fcntl
import os
import stat

from gemato.exceptions import (
    ManifestCrossDevice,
    ManifestInvalidPath,
    ManifestNoSupportedHashes,
    ManifestInsecureHashes,
    )
from gemato.hash import hash_file
from gemato.manifest import (
    manifest_hashes_to_hashlib,
    is_hash_supported,
    is_hash_secure,
    )


def get_file_metadata(path, hashes):
    """
    Get a generator for the metadata of the file at system path @path.

    The generator yields, in order:
    1. A boolean indicating whether the file exists.
    2. st_dev, if the file exists.
    3. Tuple of (S_IFMT(st_mode), file type as string), if the file
       exists.
    4. st_size, if the file exists and is a regular file. Note that
       it may be 0 on some filesystems, so treat the value with caution.
    5. st_mtime, if the file exists and is a regular file.
    6. A dict of @hashes and their values, if the file exists and is
       a regular file. Special __size__ member is added unconditionally.

    Note that the generator acquires resources, and does not release
    them until terminated. Always make sure to pull it until
    StopIteration, or close it explicitly.
    """

    open_exc = None
    try:
        # we want O_NONBLOCK to avoid blocking when opening pipes
        fd = os.open(path, os.O_RDONLY | os.O_NONBLOCK)
    except FileNotFoundError:
        exists = False
        opened = False
    except OSError as err:
        if err.errno in (errno.ENXIO, errno.EOPNOTSUPP):
            # ENXIO = unconnected device or socket
            # EOPNOTSUPP = opening UNIX socket on FreeBSD
            exists = True
            opened = False
            open_exc = err
        else:
            raise
    else:
        exists = True
        opened = True

    try:
        # 1. does it exist?
        yield exists

        # we can't provide any more data for a file that does not exist
        if not exists:
            return

        if opened:
            st = os.fstat(fd)
        else:
            st = os.stat(path)

        # safety check: if open() failed, it should not be a regular file
        if not opened and stat.S_ISREG(st.st_mode):
            raise open_exc

        # 2. st_dev
        yield st.st_dev

        # 3. file type tuple
        if stat.S_ISREG(st.st_mode):
            ftype = 'regular file'
        elif stat.S_ISDIR(st.st_mode):
            ftype = 'directory'
        elif stat.S_ISCHR(st.st_mode):
            ftype = 'character device'
        elif stat.S_ISBLK(st.st_mode):
            ftype = 'block device'
        elif stat.S_ISFIFO(st.st_mode):
            ftype = 'named pipe'
        elif stat.S_ISSOCK(st.st_mode):
            ftype = 'UNIX socket'
        else:
            ftype = 'unknown'
        yield (stat.S_IFMT(st.st_mode), ftype)

        if not stat.S_ISREG(st.st_mode):
            if opened:
                os.close(fd)
            return

        # 4. st_size
        yield st.st_size

        # 5. st_mtime
        yield st.st_mtime

        f = open(fd, 'rb')
    except Exception:
        if opened:
            os.close(fd)
        raise

    with f:
        # open() might have left the file as O_NONBLOCK
        # make sure to fix that
        fcntl.fcntl(fd, fcntl.F_SETFL, 0)

        # 5. checksums
        e_hashes = sorted(hashes)
        hashes = list(manifest_hashes_to_hashlib(e_hashes))
        e_hashes.append('__size__')
        hashes.append('__size__')
        checksums = hash_file(f, hashes, _apparent_size=st.st_size)

        ret = {}
        for ek, k in zip(e_hashes, hashes):
            ret[ek] = checksums[k]
        yield ret


def verify_path(path, e, expected_dev=None, last_mtime=None,
                require_secure_hash=False):
    """
    Verify the file at system path @path against the data in entry @e.
    The path/filename is not matched against the entry -- the correct
    entry must be passed by the caller.

    If the path passes verification, returns (True, []). Otherwise,
    returns (False, diff) where diff is a list of differences between
    the file at path and the Manifest entry. Each list element is
    a tuple of (name, expected, got).

    If @expected_dev is not None, verifies that the file resides
    on specified device. If the device does not match, raises
    ManifestCrossDevice exception. It can be used to verify that
    the files do not cross filesystem boundaries.

    If @last_mtime is not None, it specifies the timestamp corresponding
    to the previous file verification. If the file is not newer
    than that, the checksum verification is skipped.

    If @require_secure_hash is True, the file must have at least one
    hash that is considered cryptographically secure.

    Each name can be:
    - __exists__ (boolean) to indicate whether the file existed,
    - __type__ (string) as a human-readable description of file type,
    - __size__ (int) as file size,
    - any checksum name according to the entry.
    """

    if e is not None:
        assert e.tag != 'TIMESTAMP'

        # IGNORE entries cause verification to always succeed
        if e.tag == 'IGNORE':
            return (True, [])

    # None indicates we have no entry, so the file must not exist
    if e is None:
        expect_exist = False
        checksums = ()
    else:
        expect_exist = True
        checksums = list(filter(is_hash_supported, e.checksums))
        if not checksums and e.checksums:
            raise ManifestNoSupportedHashes(path, e)
        if require_secure_hash:
            # Note: even if we require secure hashes, we verify all of them
            secure_hashes = list(filter(is_hash_secure, checksums))
            if not secure_hashes:
                raise ManifestInsecureHashes(checksums)

    with contextlib.closing(get_file_metadata(path, checksums)) as g:
        # 1. verify whether the file existed in the first place
        exists = next(g)
        if exists != expect_exist:
            return (False, [('__exists__', expect_exist, exists)])
        elif not exists:
            return (True, [])

        # 2. check for xdev condition
        st_dev = next(g)
        if expected_dev is not None and st_dev != expected_dev:
            raise ManifestCrossDevice(path)

        # 3. verify whether the file is a regular file
        ifmt, ftype = next(g)
        if not stat.S_ISREG(ifmt):
            return (False, [('__type__', 'regular file', ftype)])

        # 4. verify the filesize, unless st_size == 0 (to account
        #    for weird filesystems)
        st_size = next(g)
        if st_size != 0 and st_size != e.size:
            return (False, [('__size__', e.size, st_size)])

        # 5. skip checksums if file has not changed since the last time
        #    (and st_size != 0 since we can't trust weird filesystems)
        st_mtime = next(g)
        if (last_mtime is not None and st_mtime <= last_mtime
                and st_size != 0):
            return (True, [])

        # 6. verify the real size from checksum data
        checksums = next(g)
        diff = []
        size = checksums.pop('__size__')
        if size != e.size:
            diff.append(('__size__', e.size, size))

        # 7. verify the checksums
        for h in sorted(checksums):
            exp = e.checksums[h]
            got = checksums[h]
            if got != exp:
                diff.append((h, exp, got))

        if diff:
            return (False, diff)

    return (True, [])


def update_entry_for_path(path, e, hashes=None, expected_dev=None,
                          last_mtime=None, require_secure_hashes=False):
    """
    Update the data in entry @e to match the current state of file
    at path @path. Uses hashes listed in @hashes (using Manifest names),
    or the current set of hashes in @e if @hashes is None.

    Returns True if anything changed, or False if the entry did
    not change.

    The file must exist and be a regular file, and the entry must be
    of DATA, MANIFEST or a derived type. The path/filename
    is not updated nor checked.

    If @expected_dev is not None, verifies that the file resides
    on specified device. If the device does not match, raises
    ManifestCrossDevice exception. It can be used to verify that
    the files do not cross filesystem boundaries.

    If @last_mtime is not None, it specifies the timestamp corresponding
    to the previous file update. If the file is not newer than that,
    the checksum calculation is skipped.

    If @require_secure_hashes is True, only secure hashes can be used.
    """

    assert e.tag not in ('IGNORE', 'TIMESTAMP')

    if hashes is None:
        hashes = list(e.checksums)
    if require_secure_hashes:
        insecure = list(filter(lambda x: not is_hash_secure(x), hashes))
        if insecure or not hashes:
            raise ManifestInsecureHashes(insecure)

    with contextlib.closing(get_file_metadata(path, hashes)) as g:
        # 1. verify whether the file existed in the first place
        exists = next(g)
        if not exists:
            raise ManifestInvalidPath(path, ('__exists__', exists))

        # 2. check for xdev condition
        st_dev = next(g)
        if expected_dev is not None and st_dev != expected_dev:
            raise ManifestCrossDevice(path)

        # 3. verify whether the file is a regular file
        ifmt, ftype = next(g)
        if not stat.S_ISREG(ifmt):
            raise ManifestInvalidPath(path, ('__type__', ftype))

        # 4. get the apparent file size
        st_size = next(g)

        # 5. skip checksums if file has not changed since the last time
        #    (and st_size makes sense)
        st_mtime = next(g)
        if (last_mtime is not None and st_mtime <= last_mtime
                and st_size != 0 and st_size == e.size):
            return False

        # 6. get the checksums and real size
        checksums = next(g)
        size = checksums.pop('__size__')
        if st_size != 0:
            assert st_size == size, (
                f'Apparent size (st_size = {st_size}) and real size '
                f'({size}) are different!')

        if e.size != size or e.checksums != checksums:
            e.size = size
            e.checksums = checksums
            return True
        return False


def verify_entry_compatibility(e1, e2):
    """
    Verify that the two entries @e1 and @e2 are compatible.

    If the entries are compatible, returns (True, diff). Otherwise,
    returns (False, diff). Here diff is a list of differences between
    @e1 and @e2. Each list element is a tuple of (name, e1, e2).

    In case of successful comparison, the diff may contain additional
    hashes that are present only in one of the entries.
    """

    # 1. compare types
    t1 = e1.tag
    t2 = e2.tag
    assert 'TIMESTAMP' not in (t1, t2)
    if t1 != t2:
        # all those tags have compatible semantics
        COMPATIBLE_TAGS = ('MANIFEST', 'DATA', 'EBUILD', 'AUX')
        if t1 not in COMPATIBLE_TAGS or t2 not in COMPATIBLE_TAGS:
            return (False, [('__type__', t1, t2)])

    # 2. compare sizes
    if e1.size != e2.size:
        return (False, [('__size__', e1.size, e2.size)])

    # 3. compare checksums
    hashes = frozenset(e1.checksums) | frozenset(e2.checksums)
    ret = True
    diff = []
    for h in sorted(hashes):
        h1 = e1.checksums.get(h)
        h2 = e2.checksums.get(h)
        if h1 != h2:
            diff.append((h, h1, h2))
            if h1 is not None and h2 is not None:
                ret = False

    return (ret, diff)



# gemato: Utility function tests
# (c) 2017-2022 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest

from gemato.util import (
    path_starts_with,
    path_inside_dir,
    )


@pytest.mark.parametrize(
    'p1,p2,expected',
    [("", "", True),
     ("foo", "", True),
     ("foo/", "", True),
     ("foo/bar", "", True),
     ("bar", "", True),
     ("bar/", "", True),
     ("bar/bar", "", True),
     ("foo", "foo", True),
     ("foo/", "foo", True),
     ("foo/bar", "foo", True),
     ("bar", "foo", False),
     ("fooo", "foo", False),
     ("foo.", "foo", False),
     ("foo", "foo/", True),
     ("foo/", "foo/", True),
     ("foo/bar", "foo/bar/", True),
     ])
def test_path_starts_with(p1, p2, expected):
    assert path_starts_with(p1, p2) is expected


@pytest.mark.parametrize(
    'p1,p2,expected',
    [("", "", False),
     ("foo", "", True),
     ("foo/", "", True),
     ("foo/bar", "", True),
     ("bar", "", True),
     ("bar/", "", True),
     ("bar/bar", "", True),
     ("foo", "foo", False),
     ("foo/", "foo", False),
     ("foo/bar", "foo", True),
     ("bar", "foo", False),
     ("fooo", "foo", False),
     ("foo.", "foo", False),
     ("foo", "foo/", False),
     ("foo/", "foo/", False),
     ("foo/bar", "foo/bar/", False),
     ])
def test_path_inside_dir(p1, p2, expected):
    assert path_inside_dir(p1, p2) is expected



# gemato: Verification tests
# (c) 2017-2022 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import contextlib
import errno
import itertools
import os
import os.path
import socket
import stat
import unittest.mock

import pytest

from gemato.exceptions import (
    ManifestInvalidPath,
    ManifestCrossDevice,
    ManifestInsecureHashes,
    )
from gemato.hash import hash_path
from gemato.manifest import new_manifest_entry
from gemato.verify import (
    get_file_metadata,
    verify_path,
    update_entry_for_path,
    verify_entry_compatibility,
    )

from tests.testutil import disallow_writes


TEST_STRING = b'The quick brown fox jumps over the lazy dog'


@pytest.fixture(scope='module')
def test_tree(tmp_path_factory):
    """Test tree with different file types needed for tests"""
    tmp_path = tmp_path_factory.mktemp('verify-')
    with open(tmp_path / 'empty-file', 'w'):
        pass
    with open(tmp_path / 'regular-file', 'wb') as f:
        f.write(TEST_STRING)
    with open(tmp_path / 'unreadable-file', 'w') as f:
        os.chmod(f.fileno(), 0)
    os.mkdir(tmp_path / 'directory')
    os.symlink('regular-file', tmp_path / 'symlink-to-file')
    os.symlink('directory', tmp_path / 'symlink-to-directory')
    os.symlink('non-existing', tmp_path / 'symlink-broken')
    os.mkfifo(tmp_path / 'named-pipe')
    unix_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    unix_sock.bind(str(tmp_path / 'unix-socket'))
    unix_sock.listen(1)
    disallow_writes(tmp_path)
    yield tmp_path
    unix_sock.close()


NONFILE_TEST_PATHS = [
    'directory',
    'symlink-to-directory',
    '/dev/null',  # character device
    'named-pipe',
    'unix-socket',
]
EMPTY_FILE_TEST_PATHS = [
    'empty-file',
    '/proc/version',  # special file
]
NONEMPTY_FILE_TEST_PATHS = [
    'regular-file',
    'symlink-to-file',
]
FILE_TEST_PATHS = EMPTY_FILE_TEST_PATHS + NONEMPTY_FILE_TEST_PATHS
NONEXIST_TEST_PATHS = [
    'non-existing',
    'symlink-broken',
]
ALL_TEST_PATHS = (NONFILE_TEST_PATHS + FILE_TEST_PATHS +
                  NONEXIST_TEST_PATHS)

TEST_PATH_TYPES = {
    'directory': (stat.S_IFDIR, 'directory'),
    'symlink-to-directory': (stat.S_IFDIR, 'directory'),
    '/dev/null': (stat.S_IFCHR, 'character device'),
    'named-pipe': (stat.S_IFIFO, 'named pipe'),
    'unix-socket': (stat.S_IFSOCK, 'UNIX socket'),
    'empty-file': (stat.S_IFREG, 'regular file'),
    'regular-file': (stat.S_IFREG, 'regular file'),
    'symlink-to-file': (stat.S_IFREG, 'regular file'),
    '/proc/version': (stat.S_IFREG, 'regular file'),
}


def get_checksums(path):
    """Get checksums for the specified path"""
    try:
        hashes = hash_path(path, ['md5', 'sha1', '__size__'])
    except FileNotFoundError:
        return None

    return {
        'MD5': hashes['md5'],
        'SHA1': hashes['sha1'],
        '__size__': hashes['__size__'],
    }


EMPTY_CHECKSUMS = {
    "MD5": "d41d8cd98f00b204e9800998ecf8427e",
    "SHA1": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
    "SHA256": "e3b0c44298fc1c149afbf4c8996fb924"
              "27ae41e4649b934ca495991b7852b855",
}
TEST_PATH_SIZES = {
    'empty-file': 0,
    'regular-file': 43,
    'symlink-to-file': 43,
    '/proc/version': 0,
}
TEST_PATH_CHECKSUMS = {
    'empty-file': {'MD5': EMPTY_CHECKSUMS["MD5"],
                   'SHA1': EMPTY_CHECKSUMS["SHA1"],
                   '__size__': TEST_PATH_SIZES['empty-file'],
                   },
    'regular-file': {'MD5': '9e107d9d372bb6826bd81d3542a419d6',
                     'SHA1': '2fd4e1c67a2d28fced849ee1bb76e7391b93eb12',
                     '__size__': TEST_PATH_SIZES['regular-file'],
                     },
    'symlink-to-file': {'MD5': '9e107d9d372bb6826bd81d3542a419d6',
                        'SHA1': '2fd4e1c67a2d28fced84'
                                '9ee1bb76e7391b93eb12',
                        '__size__': TEST_PATH_SIZES['regular-file'],
                        },
    '/proc/version': get_checksums('/proc/version'),
}


def strip_size(checksums):
    """Strip __size__ from checksum dict"""
    d = dict(checksums)
    del d['__size__']
    return d


@pytest.mark.parametrize(
    # note: None for dynamic components, test will fill the blanks in
    'path,expected',
    [(path, [False]) for path in NONEXIST_TEST_PATHS] +
    [(path, [True, None, TEST_PATH_TYPES[path]])
     for path in NONFILE_TEST_PATHS] +
    [(path, [True, None, TEST_PATH_TYPES[path], TEST_PATH_SIZES[path],
             None, TEST_PATH_CHECKSUMS[path]])
     for path in FILE_TEST_PATHS])
def test_get_file_metadata(test_tree, path, expected):
    if path.startswith('/') and not os.path.exists(path):
        pytest.skip(f'{path} does not exist')

    try:
        st = os.stat(test_tree / path)
    except FileNotFoundError:
        pass
    else:
        # fill in the blanks
        if len(expected) > 1:
            assert expected[1] is None
            expected[1] = st.st_dev
        if len(expected) > 4:
            assert expected[4] is None
            expected[4] = st.st_mtime

    assert (
        list(get_file_metadata(test_tree / path,
                               hashes=['MD5', 'SHA1'])) ==
        expected)


EMPTY_FILE_DATA = [0, {}]
ZERO_MD5 = '00000000000000000000000000000000'
ZERO_SHA1 = '0000000000000000000000000000000000000000'


def mangle_one_checksum(checksums):
    """Returns checksums with MD5 mangled"""
    d = strip_size(checksums)
    d['MD5'] = ZERO_MD5
    if checksums['MD5'] == d['MD5']:
        pytest.skip('MD5 already was zeros, how likely is that!?')
    return d


def mangle_both_checksums(checksums):
    """Returns checksums with MD5 and SHA1 mangled"""
    d = strip_size(checksums)
    d['MD5'] = ZERO_MD5
    d['SHA1'] = ZERO_SHA1
    if checksums['MD5'] == d['MD5']:
        pytest.skip('MD5 already was zeros, how likely is that!?')
    if checksums['SHA1'] == d['SHA1']:
        pytest.skip('SHA1 already was zeros, how likely is that!?')
    return d


class FILE_MTIME:
    pass


@pytest.mark.parametrize(
    'path,entry,args,last_mtime,expected,diff',
    # IGNORE should pass for everything, even unreadable
    [(path, 'IGNORE', [], None, True, [])
     for path in ALL_TEST_PATHS + ['unreadable-file']] +
    # None means must not exist, so passes for non-existing,
    # fails for everything existing
    [(path, None, [], None, True, []) for path in NONEXIST_TEST_PATHS] +
    [(path, None, [], None, False, [('__exists__', False, True)])
     for path in ALL_TEST_PATHS if path not in NONEXIST_TEST_PATHS] +
    # test DATA on non-regular files
    [('non-existing', 'DATA', EMPTY_FILE_DATA,
      None, False, [('__exists__', True, False)])] +
    [(path, 'DATA', EMPTY_FILE_DATA,
      None, False,
      [('__type__', 'regular file', TEST_PATH_TYPES[path][1])])
     for path in NONFILE_TEST_PATHS] +
    # test DATA on regular files
    list(itertools.chain.from_iterable(
        [(path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'], {}],
          None, True, []),
         (path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'],
                         strip_size(TEST_PATH_CHECKSUMS[path])],
          None, True, []),
         # wrong size
         (path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'] + 11, {}],
          None, False, [('__size__',
                         TEST_PATH_CHECKSUMS[path]['__size__'] + 11,
                         TEST_PATH_CHECKSUMS[path]['__size__'])]),
         # one wrong checksum
         (path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'],
                         mangle_one_checksum(TEST_PATH_CHECKSUMS[path])],
          None, False,
          [('MD5', ZERO_MD5, TEST_PATH_CHECKSUMS[path]['MD5'])]),
         # both wrong checksums
         (path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'],
                         mangle_both_checksums(
                             TEST_PATH_CHECKSUMS[path])],
          None, False,
          [('MD5', ZERO_MD5, TEST_PATH_CHECKSUMS[path]['MD5']),
           ('SHA1', ZERO_SHA1, TEST_PATH_CHECKSUMS[path]['SHA1'])]),
         ] for path in FILE_TEST_PATHS)) +
    # both wrong checksums + size (different for st_size == 0)
    [(path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'] + 11,
                     mangle_both_checksums(
                         TEST_PATH_CHECKSUMS[path])],
      None, False, [('__size__',
                     TEST_PATH_CHECKSUMS[path]['__size__'] + 11,
                     TEST_PATH_CHECKSUMS[path]['__size__'])])
     for path in NONEMPTY_FILE_TEST_PATHS] +
    # on empty files, mtime does not matter
    [(path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'],
                     mangle_both_checksums(TEST_PATH_CHECKSUMS[path])],
      mtime, False,
      [('MD5', ZERO_MD5, TEST_PATH_CHECKSUMS[path]['MD5']),
       ('SHA1', ZERO_SHA1, TEST_PATH_CHECKSUMS[path]['SHA1'])])
     for path in EMPTY_FILE_TEST_PATHS
     for mtime in (0, FILE_MTIME)] +
    # on non-empty files with correct size, up-to-date mtime skips
    # checksum
    list(itertools.chain.from_iterable(
        [(path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'],
                         mangle_both_checksums(TEST_PATH_CHECKSUMS[path])],
          0, False,
          [('MD5', ZERO_MD5, TEST_PATH_CHECKSUMS[path]['MD5']),
           ('SHA1', ZERO_SHA1, TEST_PATH_CHECKSUMS[path]['SHA1'])]),
         (path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'],
                         mangle_both_checksums(TEST_PATH_CHECKSUMS[path])],
          FILE_MTIME, True, []),
         # but size change invalidates it
         (path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'] + 11,
                         mangle_both_checksums(TEST_PATH_CHECKSUMS[path])],
          FILE_MTIME, False,
          [('__size__',
            TEST_PATH_CHECKSUMS[path]['__size__'] + 11,
            TEST_PATH_CHECKSUMS[path]['__size__'])]),
         ] for path in NONEMPTY_FILE_TEST_PATHS)) +
    [(path, None, [], FILE_MTIME, False, [('__exists__', False, True)])
     for path in FILE_TEST_PATHS])
def test_verify_path(test_tree, path, entry, args, last_mtime, expected,
                     diff):
    if path.startswith('/') and not os.path.exists(path):
        pytest.skip(f'{path} does not exist')
    if entry is not None:
        entry = new_manifest_entry(entry, path, *args)
    if last_mtime is FILE_MTIME:
        st = os.stat(test_tree / path)
        last_mtime = st.st_mtime
    assert verify_path(test_tree / path,
                       entry,
                       last_mtime=last_mtime) == (expected, diff)


@pytest.mark.parametrize(
    'path,key,match',
    [(path, '__exists__', False)
     for path in NONEXIST_TEST_PATHS] +
    [(path, '__type__', TEST_PATH_TYPES[path][1])
     for path in NONFILE_TEST_PATHS])
def test_update_fail(test_tree, path, key, match):
    entry = new_manifest_entry('DATA', path, 0, {})
    with pytest.raises(ManifestInvalidPath) as exc:
        update_entry_for_path(test_tree / path, entry)
    assert exc.value.detail == (key, match)


@pytest.mark.parametrize('function', [verify_path,
                                      update_entry_for_path])
def test_cross_filesystem(test_tree, function):
    filename = 'empty-file'
    try:
        st = os.stat('/proc')
        lst = os.stat(test_tree / filename)
    except OSError:
        pytest.skip('unable to stat /proc or empty-file')
    if st.st_dev == lst.st_dev:
        pytest.skip('/proc and test tree on the same filesystem!?')

    entry = new_manifest_entry('DATA', filename, 0, {})
    with pytest.raises(ManifestCrossDevice):
        function(test_tree / filename, entry, expected_dev=st.st_dev)


@pytest.mark.parametrize(
    'path,cls,args,new_hashes,last_mtime,retval,new_data',
    list(itertools.chain.from_iterable(
        [(path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'], {}],
          None, None, False,
          [('size', TEST_PATH_CHECKSUMS[path]['__size__']),
           ]),
         (path, 'MISC', [TEST_PATH_CHECKSUMS[path]['__size__'], {}],
          None, None, False,
          [('size', TEST_PATH_CHECKSUMS[path]['__size__']),
           ]),
         # unchanged hashes
         (path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'],
                         strip_size(TEST_PATH_CHECKSUMS[path])],
          None, None, False,
          [('size', TEST_PATH_CHECKSUMS[path]['__size__']),
           ('checksums', strip_size(TEST_PATH_CHECKSUMS[path])),
           ]),
         # new hashes
         (path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'], {}],
          ['MD5', 'SHA1'], None, True,
          [('size', TEST_PATH_CHECKSUMS[path]['__size__']),
           ('checksums', strip_size(TEST_PATH_CHECKSUMS[path])),
           ]),
         # fill hashes already in Manifest
         (path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'],
                         {'MD5': '', 'SHA1': ''}],
          None, None, True,
          [('size', TEST_PATH_CHECKSUMS[path]['__size__']),
           ('checksums', strip_size(TEST_PATH_CHECKSUMS[path])),
           ]),
         # subset of existing hashes
         (path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'],
                         strip_size(TEST_PATH_CHECKSUMS[path])],
          ['MD5'], None, True,
          [('size', TEST_PATH_CHECKSUMS[path]['__size__']),
           ('checksums', {'MD5': TEST_PATH_CHECKSUMS[path]['MD5']}),
           ]),
         # superset of existing hashes
         (path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'],
                         {'MD5': TEST_PATH_CHECKSUMS[path]['MD5']}],
          ['MD5', 'SHA1'], None, True,
          [('size', TEST_PATH_CHECKSUMS[path]['__size__']),
           ('checksums', strip_size(TEST_PATH_CHECKSUMS[path])),
           ]),
         ] for path in FILE_TEST_PATHS)) +
    # mtime does not affect empty files
    [(path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'],
                     {'MD5': ZERO_MD5, 'SHA1': ZERO_SHA1}],
      ['MD5', 'SHA1'], mtime, True,
      [('size', TEST_PATH_CHECKSUMS[path]['__size__']),
       ('checksums', strip_size(TEST_PATH_CHECKSUMS[path])),
       ])
     for path in EMPTY_FILE_TEST_PATHS
     for mtime in (0, FILE_MTIME)] +
    # but non-empty files with recent mtime do not get rechecked
    list(itertools.chain.from_iterable(
        [(path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'],
                         {'MD5': ZERO_MD5, 'SHA1': ZERO_SHA1}],
          ['MD5', 'SHA1'], 0, True,
          [('size', TEST_PATH_CHECKSUMS[path]['__size__']),
           ('checksums', strip_size(TEST_PATH_CHECKSUMS[path])),
           ]),
         (path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'],
                         {'MD5': ZERO_MD5, 'SHA1': ZERO_SHA1}],
          ['MD5', 'SHA1'], FILE_MTIME, False,
          [('size', TEST_PATH_CHECKSUMS[path]['__size__']),
           ('checksums', {'MD5': ZERO_MD5, 'SHA1': ZERO_SHA1}),
           ]),
         # size invalidates recent mtime
         (path, 'DATA', [TEST_PATH_CHECKSUMS[path]['__size__'] + 11,
                         {'MD5': ZERO_MD5, 'SHA1': ZERO_SHA1}],
          ['MD5', 'SHA1'], FILE_MTIME, True,
          [('size', TEST_PATH_CHECKSUMS[path]['__size__']),
           ('checksums', strip_size(TEST_PATH_CHECKSUMS[path])),
           ]),
         ] for path in NONEMPTY_FILE_TEST_PATHS)) +
    [])
def test_update(test_tree, path, cls, args, new_hashes, last_mtime,
                retval, new_data):
    entry = new_manifest_entry(cls, path, *args)
    if last_mtime is FILE_MTIME:
        st = os.stat(test_tree / path)
        last_mtime = st.st_mtime
    assert update_entry_for_path(test_tree / path,
                                 entry,
                                 hashes=new_hashes,
                                 last_mtime=last_mtime) is retval
    assert entry.path == path
    for k, v in new_data:
        assert getattr(entry, k) == v


def test_update_IGNORE(test_tree):
    path = 'empty-file'
    entry = new_manifest_entry('IGNORE', path)
    with pytest.raises(AssertionError):
        update_entry_for_path(test_tree / path, entry)


def test_update_AUX(test_tree):
    path = 'empty-file'
    entry = new_manifest_entry('AUX', path, *EMPTY_FILE_DATA)
    assert not update_entry_for_path(test_tree / path, entry)
    assert entry.aux_path == path
    assert entry.path == f'files/{path}'
    assert entry.size == 0
    assert entry.checksums == {}


@pytest.mark.parametrize(
    'function,args',
    [(get_file_metadata, [[]]),
     (verify_path,
      [new_manifest_entry('DATA', 'unreadable-file', 0, {})]),
     (update_entry_for_path,
      [new_manifest_entry('DATA', 'unreadable-file', 0, {})]),
     ])
def test_unreadable_file(test_tree, function, args):
    with pytest.raises(PermissionError):
        for x in function(test_tree / 'unreadable-file', *args):
            pass


@pytest.mark.parametrize(
    "entry_hash,hashes_arg,insecure",
    [("MD5", None, True),
     ("SHA1", None, True),
     ("SHA512", None, False),
     ("MD5", "SHA1 SHA512", True),
     ("MD5", "SHA512", False),
     ("SHA512", "MD5 SHA512", True),
     ("", None, True),
     ("", "SHA512", False),
     ("SHA512", "", True),
     ])
def test_insecure_hashes(test_tree, entry_hash, hashes_arg, insecure):
    ctx = (pytest.raises(ManifestInsecureHashes) if insecure
           else contextlib.nullcontext())
    with ctx:
        update_entry_for_path(
            test_tree / "empty-file",
            new_manifest_entry("DATA", "empty-file", 0,
                               {entry_hash: ""} if entry_hash else {}),
            hashes=hashes_arg.split() if hashes_arg is not None else None,
            require_secure_hashes=True)


@pytest.mark.parametrize(
    "entry_hash,insecure",
    [("", True),
     ("MD5", True),
     ("SHA1", True),
     ("SHA256", False),
     ("MD5 SHA1", True),
     ("SHA1 SHA256", False),
     ])
def test_verify_insecure_hashes(test_tree, entry_hash, insecure):
    checksums = {}
    for h in entry_hash.split():
        checksums[h] = EMPTY_CHECKSUMS[h]

    ctx = (pytest.raises(ManifestInsecureHashes) if insecure
           else contextlib.nullcontext())
    with ctx:
        verify_path(
            test_tree / "empty-file",
            new_manifest_entry("DATA", "empty-file", 0, checksums),
            require_secure_hash=True)


@pytest.mark.parametrize(
    'a_cls,a_name,a_args,b_cls,b_name,b_args,expected,diff',
    [('DATA', 'test', [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      'DATA', 'test', [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      True, []),
     ('DATA', 'test-1.ebuild',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      'EBUILD', 'test-1.ebuild',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      True, []),
     ('DATA', 'files/test.patch',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      'AUX', 'test.patch',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      True, []),
     ('DATA', 'Manifest',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      'MANIFEST', 'Manifest',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      True, []),
     ('DATA', 'metadata.xml',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      'MISC', 'metadata.xml',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      False, [('__type__', 'DATA', 'MISC')]),
     ('DATA', 'test',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      'IGNORE', 'test', [],
      False, [('__type__', 'DATA', 'IGNORE')]),
     ('DATA', 'test-1.tar.gz',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      'DIST', 'test-1.tar.gz',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      False, [('__type__', 'DATA', 'DIST')]),
     ('DATA', 'mismatched-size',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      'DATA', 'mismatched-size',
      [32, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      False, [('__size__', 0, 32)]),
     ('DATA', 'mismatched-md5',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      'DATA', 'mismatched-md5',
      [0, {'MD5': ZERO_MD5}],
      False, [('MD5', 'd41d8cd98f00b204e9800998ecf8427e', ZERO_MD5)]),
     ('DATA', 'hash-subset',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      'DATA', 'mismatched-md5',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e',
           'SHA1': 'da39a3ee5e6b4b0d3255bfef95601890afd80709'}],
      True, [('SHA1', None, 'da39a3ee5e6b4b0d3255bfef95601890afd80709')]),
     ('DATA', 'mismatched-hash-sets',
      [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
      'DATA', 'mismatched-md5',
      [0, {'SHA1': 'da39a3ee5e6b4b0d3255bfef95601890afd80709'}],
      True, [('MD5', 'd41d8cd98f00b204e9800998ecf8427e', None),
             ('SHA1', None, 'da39a3ee5e6b4b0d3255bfef95601890afd80709')]),
     ])
def test_entry_compatibility(a_cls, a_name, a_args, b_cls, b_name,
                             b_args, expected, diff):
    e1 = new_manifest_entry(a_cls, a_name, *a_args)
    e2 = new_manifest_entry(b_cls, b_name, *b_args)
    assert verify_entry_compatibility(e1, e2) == (expected, diff)


def test_get_file_metadata_fail_to_open_reg(test_tree):
    """Regression test for when open() fails on a seemingly regular file"""
    with unittest.mock.patch("os.open") as mock_open:
        mock_open.side_effect = OSError(errno.ENXIO, "mocked error")
        with pytest.raises(OSError):
            list(get_file_metadata(test_tree / "regular-file", {}))



# gemato: Manifest file support tests
# (c) 2017-2022 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import datetime
import io

import pytest

from gemato.exceptions import (
    ManifestSyntaxError,
    )
from gemato.manifest import (
    ManifestFile,
    ManifestEntryTIMESTAMP,
    ManifestEntryMANIFEST,
    ManifestEntryIGNORE,
    ManifestEntryDATA,
    ManifestEntryMISC,
    ManifestEntryDIST,
    ManifestEntryEBUILD,
    ManifestEntryAUX,
    manifest_hashes_to_hashlib,
    new_manifest_entry,
    )


TEST_MANIFEST = '''
TIMESTAMP 2017-10-22T18:06:41Z
MANIFEST eclass/Manifest 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
 SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
IGNORE local
DATA myebuild-0.ebuild 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
 SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
MISC metadata.xml 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
 SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
DIST mydistfile.tar.gz 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
 SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
DATA foo.txt 0
'''

TEST_DEPRECATED_MANIFEST = '''
EBUILD myebuild-0.ebuild 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
 SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
MISC metadata.xml 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
 SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
AUX test.patch 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
 SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
DIST mydistfile.tar.gz 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
 SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
'''

EMPTY_MANIFEST = ''


@pytest.mark.parametrize('manifest_var', ['TEST_MANIFEST',
                                          'TEST_DEPRECATED_MANIFEST',
                                          'EMPTY_MANIFEST'])
def test_load(manifest_var):
    with io.StringIO(globals()[manifest_var]) as f:
        m = ManifestFile()
        m.load(f)


@pytest.mark.parametrize('manifest_var', ['TEST_MANIFEST',
                                          'TEST_DEPRECATED_MANIFEST',
                                          'EMPTY_MANIFEST'])
def test_load_via_ctor(manifest_var):
    with io.StringIO(globals()[manifest_var]) as f:
        ManifestFile(f)


@pytest.mark.parametrize('manifest_var', ['TEST_MANIFEST',
                                          'TEST_DEPRECATED_MANIFEST',
                                          'EMPTY_MANIFEST'])
def test_load_and_dump(manifest_var):
    m = ManifestFile()
    with io.StringIO(globals()[manifest_var]) as f:
        m.load(f)
    with io.StringIO() as outf:
        m.dump(outf)
        assert outf.getvalue().strip() == globals()[manifest_var].strip()


@pytest.mark.parametrize(
    'manifest_var,expected',
    [('TEST_MANIFEST', datetime.datetime(2017, 10, 22, 18, 6, 41)),
     ('TEST_DEPRECATED_MANIFEST', None),
     ('EMPTY_MANIFEST', None),
     ])
def test_find_timestamp(manifest_var, expected):
    m = ManifestFile()
    with io.StringIO(globals()[manifest_var]) as f:
        m.load(f)
    if expected is None:
        assert m.find_timestamp() is None
    else:
        assert m.find_timestamp().ts == expected


@pytest.fixture(scope='module')
def test_manifest():
    m = ManifestFile()
    with io.StringIO(TEST_MANIFEST) as f:
        m.load(f)
    yield m


@pytest.fixture(scope='module')
def deprecated_manifest():
    m = ManifestFile()
    with io.StringIO(TEST_DEPRECATED_MANIFEST) as f:
        m.load(f)
    yield m


@pytest.mark.parametrize(
    'path,expected',
    [('2017-10-22T18:06:41Z', None),
     ('eclass/Manifest', 'eclass/Manifest'),
     ('eclass', None),
     ('local', 'local'),
     ('local/foo', 'local'),
     ('locale', None),
     ('myebuild-0.ebuild', 'myebuild-0.ebuild'),
     ('metadata.xml', 'metadata.xml'),
     ('mydistfile.tar.gz', None),
     ])
def test_find_path_entry(test_manifest, path, expected):
    pe = test_manifest.find_path_entry(path)
    if expected is None:
        assert pe is None
    else:
        assert pe.path == expected


@pytest.mark.parametrize(
    'path,expected',
    [('test.patch', None),
     ('files/test.patch', 'test.patch'),
     ])
def test_find_path_entry_aux(deprecated_manifest, path, expected):
    pe = deprecated_manifest.find_path_entry(path)
    if expected is None:
        assert pe is None
    else:
        assert pe.aux_path == expected


@pytest.mark.parametrize(
    'filename,expected',
    [('myebuild-0.ebuild', None),
     ('mydistfile.tar.gz', 'mydistfile.tar.gz'),
     ])
def test_find_dist_entry(test_manifest, filename, expected):
    pe = test_manifest.find_dist_entry(filename)
    if expected is None:
        assert pe is None
    else:
        assert pe.path == expected


@pytest.mark.parametrize(
    'path,expected',
    [('foo', []),
     ('eclass', []),
     ('eclass/foo.eclass', ['eclass/Manifest']),
     ])
def test_find_manifests_for_path(test_manifest, path, expected):
    assert [x.path for x
            in test_manifest.find_manifests_for_path(path)] == expected


def test_multiple_load():
    """Test that load() overwrites previously loaded data."""
    m = ManifestFile()
    with io.StringIO(TEST_MANIFEST) as f:
        m.load(f)
    with io.StringIO(TEST_DEPRECATED_MANIFEST) as f:
        m.load(f)
    with io.StringIO() as outf:
        m.dump(outf)
        assert outf.getvalue().strip() == TEST_DEPRECATED_MANIFEST.strip()


def test_sort():
    m = ManifestFile()
    with io.StringIO(TEST_MANIFEST) as f:
        m.load(f)
    with io.StringIO() as outf:
        m.dump(outf, sort=True)
        assert outf.getvalue().strip() == '''
DATA foo.txt 0
DATA myebuild-0.ebuild 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
 SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
DIST mydistfile.tar.gz 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
 SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
IGNORE local
MANIFEST eclass/Manifest 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
 SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
MISC metadata.xml 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
 SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
TIMESTAMP 2017-10-22T18:06:41Z
'''.strip()


TEST_FILE_DATA = [('path', 'test'),
                  ('size', 0),
                  ('checksums', {'MD5': 'd41d8cd98f00b204'
                                        'e9800998ecf8427e',
                                 'SHA1': 'da39a3ee5e6b4b0d3255'
                                         'bfef95601890afd80709',
                                 }),
                  ]
TEST_FILE_LIST = ['test', '0', 'MD5', 'd41d8cd98f00b204e9800998ecf8427e',
                  'SHA1', 'da39a3ee5e6b4b0d3255bfef95601890afd80709']


ENTRY_TEST_DATA = [
    (ManifestEntryTIMESTAMP,
     ['TIMESTAMP', '2010-01-01T11:12:13Z'],
     [('ts', datetime.datetime(2010, 1, 1, 11, 12, 13)),
      ]),
    (ManifestEntryMANIFEST,
     ['MANIFEST'] + TEST_FILE_LIST,
     TEST_FILE_DATA),
    (ManifestEntryIGNORE,
     ['IGNORE', 'test'],
     [('path', 'test'),
      ]),
    (ManifestEntryDATA,
     ['DATA'] + TEST_FILE_LIST,
     TEST_FILE_DATA),
    (ManifestEntryIGNORE,
     ['IGNORE', 'test'],
     [('path', 'test'),
      ]),
    (ManifestEntryMISC,
     ['MISC'] + TEST_FILE_LIST,
     TEST_FILE_DATA),
    (ManifestEntryDIST,
     ['DIST'] + TEST_FILE_LIST,
     TEST_FILE_DATA),
    (ManifestEntryEBUILD,
     ['EBUILD'] + TEST_FILE_LIST,
     TEST_FILE_DATA),
    (ManifestEntryAUX,
     ['AUX'] + TEST_FILE_LIST,
     TEST_FILE_DATA),
]


@pytest.mark.parametrize('cls,as_list,vals', ENTRY_TEST_DATA)
def test_entry_from_list(cls, as_list, vals):
    entry = cls.from_list(as_list)
    for k, v in vals:
        if cls is ManifestEntryAUX and k == 'path':
            assert entry.aux_path == v
            v = 'files/' + v
        assert getattr(entry, k) == v


@pytest.mark.parametrize('cls,as_list,vals', ENTRY_TEST_DATA)
def test_entry_to_list(cls, as_list, vals):
    entry = cls(*(v for k, v in vals))
    assert list(entry.to_list()) == as_list


@pytest.mark.parametrize('cls,as_list,vals', ENTRY_TEST_DATA)
def test_new_manifest_entry(cls, as_list, vals):
    entry = new_manifest_entry(as_list[0], *(v for k, v in vals))
    assert isinstance(entry, cls)
    assert list(entry.to_list()) == as_list


ENTRY_INVALID_DATA = [
    (ManifestEntryTIMESTAMP, ('TIMESTAMP',)),
    (ManifestEntryTIMESTAMP, ('TIMESTAMP', '')),
    (ManifestEntryTIMESTAMP, ('TIMESTAMP', '2017-10-22T18:06:41+02:00')),
    (ManifestEntryTIMESTAMP, ('TIMESTAMP', '2017-10-22T18:06:41')),
    (ManifestEntryTIMESTAMP, ('TIMESTAMP', '2017-10-22', '18:06:41Z')),
    (ManifestEntryTIMESTAMP, ('TIMESTAMP', '20171022T180641Z')),
    (ManifestEntryMANIFEST, ('MANIFEST', '', '0')),
    (ManifestEntryMANIFEST, ('MANIFEST', '/foo', '0')),
    (ManifestEntryIGNORE, ('IGNORE', '',)),
    (ManifestEntryIGNORE, ('IGNORE', '/foo',)),
    (ManifestEntryDATA, ('DATA', '',)),
    (ManifestEntryDATA, ('DATA', '/foo',)),
    (ManifestEntryMISC, ('MISC', '',)),
    (ManifestEntryMISC, ('MISC', '/foo',)),
    (ManifestEntryDIST, ('DIST', '',)),
    (ManifestEntryDIST, ('DIST', '/foo',)),
    (ManifestEntryDIST, ('DIST', 'foo/bar.gz',)),
    (ManifestEntryEBUILD, ('EBUILD', '',)),
    (ManifestEntryEBUILD, ('EBUILD', '/foo',)),
    (ManifestEntryAUX, ('AUX', '',)),
    (ManifestEntryAUX, ('AUX', '/foo',)),
    (ManifestEntryMANIFEST, ('MANIFEST', 'foo', 'asdf')),
    (ManifestEntryMANIFEST, ('MANIFEST', 'foo', '5ds')),
    (ManifestEntryMANIFEST, ('MANIFEST', 'foo', '-5')),
    (ManifestEntryDATA, ('DATA', 'foo', 'asdf')),
    (ManifestEntryDATA, ('DATA', 'foo', '5ds')),
    (ManifestEntryDATA, ('DATA', 'foo', '-5')),
    (ManifestEntryMISC, ('MISC', 'foo', 'asdf')),
    (ManifestEntryMISC, ('MISC', 'foo', '5ds')),
    (ManifestEntryMISC, ('MISC', 'foo', '-5')),
    (ManifestEntryDIST, ('DIST', 'foo', 'asdf')),
    (ManifestEntryDIST, ('DIST', 'foo', '5ds')),
    (ManifestEntryDIST, ('DIST', 'foo', '-5')),
    (ManifestEntryEBUILD, ('EBUILD', 'foo', 'asdf')),
    (ManifestEntryEBUILD, ('EBUILD', 'foo', '5ds')),
    (ManifestEntryEBUILD, ('EBUILD', 'foo', '-5')),
    (ManifestEntryAUX, ('AUX', 'foo', 'asdf')),
    (ManifestEntryAUX, ('AUX', 'foo', '5ds')),
    (ManifestEntryAUX, ('AUX', 'foo', '-5')),
    (ManifestEntryMANIFEST, ('MANIFEST', 'foo', '0', 'md5')),
    (ManifestEntryMANIFEST, ('MANIFEST', 'foo', '0', 'md5',
                             'd41d8cd98f00b204e9800998ecf8427e', 'sha1')),
    (ManifestEntryDATA, ('DATA', 'foo', '0', 'md5')),
    (ManifestEntryDATA, ('DATA', 'foo', '0', 'md5',
                         'd41d8cd98f00b204e9800998ecf8427e', 'sha1')),
    (ManifestEntryMISC, ('MISC', 'foo', '0', 'md5')),
    (ManifestEntryMISC, ('MISC', 'foo', '0', 'md5',
                         'd41d8cd98f00b204e9800998ecf8427e', 'sha1')),
    (ManifestEntryDIST, ('DIST', 'foo', '0', 'md5')),
    (ManifestEntryDIST, ('DIST', 'foo', '0', 'md5',
                         'd41d8cd98f00b204e9800998ecf8427e', 'sha1')),
    (ManifestEntryEBUILD, ('EBUILD', 'foo', '0', 'md5')),
    (ManifestEntryEBUILD, ('EBUILD', 'foo', '0', 'md5',
                           'd41d8cd98f00b204e9800998ecf8427e', 'sha1')),
    (ManifestEntryAUX, ('AUX', 'foo', '0', 'md5')),
    (ManifestEntryAUX, ('AUX', 'foo', '0', 'md5',
                        'd41d8cd98f00b204e9800998ecf8427e', 'sha1')),
    (ManifestEntryMANIFEST, ('MANIFEST',)),
    (ManifestEntryMANIFEST, ('MANIFEST', 'foo')),
    (ManifestEntryIGNORE, ('IGNORE',)),
    (ManifestEntryIGNORE, ('IGNORE', 'foo', 'bar')),
    (ManifestEntryDATA, ('DATA',)),
    (ManifestEntryDATA, ('DATA', 'foo')),
    (ManifestEntryMISC, ('MISC',)),
    (ManifestEntryMISC, ('MISC', 'foo')),
    (ManifestEntryDIST, ('DIST',)),
    (ManifestEntryDIST, ('DIST', 'foo')),
    (ManifestEntryEBUILD, ('EBUILD',)),
    (ManifestEntryEBUILD, ('EBUILD', 'foo')),
    (ManifestEntryAUX, ('AUX',)),
    (ManifestEntryAUX, ('AUX', 'foo')),
]


@pytest.mark.parametrize('cls,as_list', ENTRY_INVALID_DATA)
def test_manifest_entry_invalid(cls, as_list):
    with pytest.raises(ManifestSyntaxError):
        cls.from_list(as_list)


@pytest.mark.parametrize('mhash,hlibhash',
                         [('MD5 SHA1', 'md5 sha1'),
                          ('RMD160', 'ripemd160'),
                          ('SHA3_256 SHA256', 'sha3_256 sha256'),
                          ])
def test_manifest_hashes_to_hashlib(mhash, hlibhash):
    assert (list(manifest_hashes_to_hashlib(mhash.split())) ==
            hlibhash.split())


PATH_ENCODING_DATA = [
    ('tes t', 'tes\\x20t'),
    ('tes\tt', 'tes\\x09t'),
    ('tes\u00a0t', 'tes\\u00A0t'),
    ('tes\u2000t', 'tes\\u2000t'),
    ('tes\x00t', 'tes\\x00t'),
    ('tes\at', 'tes\\x07t'),
    ('tes\x7ft', 'tes\\x7Ft'),
    ('tes\u0080t', 'tes\\u0080t'),
    ('tes\\t', 'tes\\x5Ct'),
]
PATH_ENCODING_NC_DATA = [
    ('tes t', 'tes\\u0020t'),
    ('tes t', 'tes\\U00000020t'),
    ('tes\u00a0t', 'tes\\u00a0t'),
    ('tes\x7ft', 'tes\\x7ft'),
    ('tes\\t', 'tes\\x5ct'),
]
PATH_ENCODING_INVALID_DATA = [
    'tes\\t',
    'tes\\\\t',
    'tes\\',
    'tes\\xt',
    'tes\\x5t',
    'tes\\ut',
    'tes\\u345t',
    'tes\\Ut',
    'tes\\U0000345t',
]


@pytest.mark.parametrize('path,enc', PATH_ENCODING_DATA)
def test_path_encode(path, enc):
    m = new_manifest_entry('DATA', path, 32, {})
    assert m.path == path
    assert m.to_list() == ['DATA', enc, '32']


@pytest.mark.parametrize('path,enc',
                         PATH_ENCODING_DATA + PATH_ENCODING_NC_DATA)
def test_path_decode(path, enc):
    m = ManifestEntryDATA.from_list(['DATA', enc, 32])
    assert m.path == path


@pytest.mark.parametrize('enc', PATH_ENCODING_INVALID_DATA)
def test_path_decode_invalid(enc):
    with pytest.raises(ManifestSyntaxError):
        ManifestEntryDATA.from_list(['DATA', enc, 32])



# gemato: compressed file tests
# (c) 2017-2022 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import base64
import io

import pytest

from gemato.compression import (
    open_compressed_file,
    open_potentially_compressed_path,
    get_potential_compressed_names,
    get_compressed_suffix_from_filename,
    )


TEST_STRING = b'The quick brown fox jumps over the lazy dog'
# we need to be specific on endianness to avoid unreliably writing BOM
UTF16_TEST_STRING = TEST_STRING.decode('ASCII').encode('utf_16_be')


COMPRESSION_ALGOS = ['gz', 'bz2', 'lzma', 'xz']

COMPRESSION_DATA = {
    'baseline': {
        None: TEST_STRING,
        'gz': b'''
H4sIACbJ8FkAAwvJSFUoLM1MzlZIKsovz1NIy69QyCrNLShWyC9LLVIoAUrnJFZVKqTkpwMA
OaNPQSsAAAA=
''',
        'bz2': b'''
QlpoOTFBWSZTWUWd7mEAAAQTgEAABAA////wIAEABTQAAAGigAAAAEBoLtBqVm1CpOmzyfUX
Aw5PHXD0304jMvvfF3JFOFCQRZ3uYQ==
''',
        'lzma': b'''
XQAAAAT//////////wAqGgiiAyVm8Ut4xaIF/y7m2dIgGq00+OId6EE2+twGabs85BA0Jwnr
s2bsGhcv//zOkAA=
''',
        'xz': b'''
/Td6WFoAAATm1rRGAgAhARwAAAAQz1jMAQAqVGhlIHF1aWNrIGJyb3duIGZveCBqdW1wcyBv
dmVyIHRoZSBsYXp5IGRvZwAAxKFK5cK4XlsAAUMrrVBuVx+2830BAAAAAARZWg==
''',
    },
    'empty': {
        None: b'',
        'gz': b'H4sIACbJ8FkAAwMAAAAAAAAAAAA=',
        'bz2': b'QlpoORdyRThQkAAAAAA=',
        'lzma': b'XQAAAAT//////////wCD//v//8AAAAA=',
        'xz': b'/Td6WFoAAATm1rRGAAAAABzfRCEftvN9AQAAAAAEWVo=',
    },
    'split': {
        None: TEST_STRING,
        'gz': b'''
H4sIACbJ8FkAAwvJSFUoLM1MzlZIKsovz1NIy69QAADidbCIFAAAAB+LCAAmyfBZAAPLKs0t
KFbIL0stUijJSFXISayqVEjJTwcAlGd4GBcAAAA=
''',
        'bz2': '''
QlpoOTFBWSZTWQgcCrAAAAITgEAABAAbabLAIABBEaDR6jT9UoAAAbUXZJ48gnMg3xdyRThQ
kAgcCrBCWmg5MUFZJlNZOxleaAAABRGAQAAm1t8wIACAUaNDRtTaSgAAAcAcViIdSEhzctM/
F3JFOFCQOxleaA==
''',
        'lzma': '''
XQAAAAT//////////wAqGgiiAyVm8Ut4xaIF/y7m2dIgGq1EvQql//X0QABdAAAABP//////
////ADUdSd6zBOkOpekGFH46zix9wE9VT65OVeV479//7uUAAA==
''',
        'xz': '''
/Td6WFoAAATm1rRGAgAhARwAAAAQz1jMAQATVGhlIHF1aWNrIGJyb3duIGZveCAAIEFC5aca
LXcAASwU+AptAx+2830BAAAAAARZWv03elhaAAAE5ta0RgIAIQEcAAAAEM9YzAEAFmp1bXBz
IG92ZXIgdGhlIGxhenkgZG9nAADjZCTmHjHqggABLxeBCEmxH7bzfQEAAAAABFla
''',
    },
}


@pytest.mark.parametrize('suffix', COMPRESSION_ALGOS)
@pytest.mark.parametrize('data_group', COMPRESSION_DATA.keys())
def test_decompress(suffix, data_group):
    data = COMPRESSION_DATA[data_group]
    with io.BytesIO(base64.b64decode(data[suffix])) as f:
        with open_compressed_file(suffix, f, "rb") as z:
            assert z.read() == data[None]


@pytest.mark.parametrize('suffix', COMPRESSION_ALGOS)
def test_round_trip(suffix):
    with io.BytesIO() as f:
        with open_compressed_file(suffix, f, 'wb') as z:
            z.write(TEST_STRING)

        f.seek(0)

        with open_compressed_file(suffix, f, 'rb') as z:
            assert z.read() == TEST_STRING


@pytest.fixture(params=COMPRESSION_ALGOS)
def test_file(tmp_path, request):
    yield tmp_path / f'test.{request.param}'


@pytest.mark.parametrize('data_group', COMPRESSION_DATA.keys())
def test_open_potentially_compressed_path(test_file, data_group):
    suffix = test_file.suffix.lstrip('.')
    with open(test_file, 'wb') as wf:
        wf.write(base64.b64decode(COMPRESSION_DATA[data_group][suffix]))

    with open_potentially_compressed_path(test_file, 'rb') as cf:
        assert cf.read() == COMPRESSION_DATA[data_group][None]


def test_open_potentially_compressed_path_write(test_file):
    with open_potentially_compressed_path(test_file, 'wb') as cf:
        cf.write(TEST_STRING)

    suffix = test_file.suffix.lstrip('.')
    with open(test_file, 'rb') as rf:
        with open_compressed_file(suffix, rf, 'rb') as z:
            assert z.read() == TEST_STRING


def test_open_potentially_compressed_path_with_encoding(test_file):
    suffix = test_file.suffix.lstrip('.')
    with open(test_file, 'wb') as wf:
        with open_compressed_file(suffix, wf, 'wb') as z:
            z.write(UTF16_TEST_STRING)

    with open_potentially_compressed_path(test_file, 'r',
                                          encoding='utf_16_be') as cf:
        assert cf.read() == TEST_STRING.decode('ASCII')


@pytest.mark.parametrize('encoding,out_var', [(None, 'TEST_STRING'),
                                              ('utf_16_be',
                                               'UTF16_TEST_STRING'),
                                              ])
def test_open_potentially_compressed_path_write_with_unicode(
        test_file, encoding, out_var):
    kwargs = {}
    if encoding is not None:
        kwargs['encoding'] = encoding
    with open_potentially_compressed_path(test_file, 'w', **kwargs) as cf:
        cf.write(TEST_STRING.decode('ASCII'))

    suffix = test_file.suffix.lstrip('.')
    with open(test_file, 'rb') as rf:
        with open_compressed_file(suffix, rf, 'rb') as z:
            assert z.read() == globals()[out_var]


def test_open_potentially_compressed_path_with_encoding_line_api(test_file):
    suffix = test_file.suffix.lstrip('.')
    with open(test_file, 'wb') as wf:
        with open_compressed_file(suffix, wf, 'wb') as z:
            z.write(UTF16_TEST_STRING)

    with open_potentially_compressed_path(test_file, 'r',
                                          encoding='utf_16_be') as cf:
        assert [x for x in cf] == [TEST_STRING.decode('ASCII')]


def test_open_potentially_compressed_path_fileno_passthrough(test_file):
    fs1 = open_potentially_compressed_path(test_file, 'w',
                                           encoding='utf_16_be')
    with fs1 as cf:
        assert ([f.fileno() for f in fs1.files] ==
                [cf.fileno() for f in fs1.files])

    fs2 = open_potentially_compressed_path(test_file, 'r',
                                           encoding='utf_16_be')
    with fs2 as cf:
        assert ([f.fileno() for f in fs2.files] ==
                [cf.fileno() for f in fs2.files])


def test_get_potential_compressed_names():
    assert (
        frozenset(get_potential_compressed_names('test')) ==
        frozenset(['test'] + [f'test.{sfx}' for sfx in COMPRESSION_ALGOS]))


@pytest.mark.parametrize('suffix', COMPRESSION_ALGOS)
def test_get_compressed_suffix_from_filename(suffix):
    assert (
        get_compressed_suffix_from_filename(f'test.{suffix}') == suffix)



# gemato: hash support tests
# (c) 2017-2022 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import io

import pytest

from gemato.exceptions import (
    UnsupportedHash,
    )
from gemato.hash import (
    get_hash_by_name,
    hash_bytes,
    hash_file,
    hash_path,
    )


EMPTY_STRING = b''
TEST_STRING = b'The quick brown fox jumps over the lazy dog'


@pytest.mark.parametrize('name', ['_invalid_name_',
                                  'blake2zzzz',
                                  'sha3_987'])
def test_get_invalid(name):
    with pytest.raises(UnsupportedHash):
        get_hash_by_name(name)


REQUIRED_TEST_HASHES = {
    'TEST_STRING': {
        'md5': '9e107d9d372bb6826bd81d3542a419d6',
        'sha1': '2fd4e1c67a2d28fced849ee1bb76e7391b93eb12',
        'sha256': 'd7a8fbb307d7809469ca9abcb0082e4f'
                  '8d5651e46d3cdb762d02d0bf37c9e592',
    },
    'EMPTY_STRING': {
        'md5': 'd41d8cd98f00b204e9800998ecf8427e',
        'sha1': 'da39a3ee5e6b4b0d3255bfef95601890afd80709',
        'sha256': 'e3b0c44298fc1c149afbf4c8996fb924'
                  '27ae41e4649b934ca495991b7852b855',
    },
}


HASH_VARIANTS = [
    # name, required, test hashes
    ('md5', True,
     {'TEST_STRING': REQUIRED_TEST_HASHES['TEST_STRING']['md5'],
      'EMPTY_STRING': REQUIRED_TEST_HASHES['EMPTY_STRING']['md5'],
      }),
    ('sha1', True,
     {'TEST_STRING': REQUIRED_TEST_HASHES['TEST_STRING']['sha1'],
      'EMPTY_STRING': REQUIRED_TEST_HASHES['EMPTY_STRING']['sha1'],
      }),
    ('sha224', True,
     {'TEST_STRING': '730e109bd7a8a32b1cb9d9a09aa2'
                     '325d2430587ddbc0c38bad911525',
      'EMPTY_STRING': 'd14a028c2a3a2bc9476102bb2882'
                      '34c415a2b01f828ea62ac5b3e42f',
      }),
    ('sha256', True,
     {'TEST_STRING': REQUIRED_TEST_HASHES['TEST_STRING']['sha256'],
      'EMPTY_STRING': REQUIRED_TEST_HASHES['EMPTY_STRING']['sha256'],
      }),
    ('sha384', True,
     {'TEST_STRING': 'ca737f1014a48f4c0b6dd43cb177b0af'
                     'd9e5169367544c494011e3317dbf9a50'
                     '9cb1e5dc1e85a941bbee3d7f2afbc9b1',
      'EMPTY_STRING': '38b060a751ac96384cd9327eb1b1e36a'
                      '21fdb71114be07434c0cc7bf63f6e1da'
                      '274edebfe76f65fbd51ad2f14898b95b',
      }),
    ('sha512', True,
     {'TEST_STRING': '07e547d9586f6a73f73fbac0435ed769'
                     '51218fb7d0c8d788a309d785436bbb64'
                     '2e93a252a954f23912547d1e8a3b5ed6'
                     'e1bfd7097821233fa0538f3db854fee6',
      'EMPTY_STRING': 'cf83e1357eefb8bdf1542850d66d8007'
                      'd620e4050b5715dc83f4a921d36ce9ce'
                      '47d0d13c5d85f2b0ff8318d2877eec2f'
                      '63b931bd47417a81a538327af927da3e',
      }),
    ('md4', False,
     {'TEST_STRING': '1bee69a46ba811185c194762abaeae90',
      'EMPTY_STRING': '31d6cfe0d16ae931b73c59d7e0c089c0',
      }),
    ('ripemd160', False,
     {'TEST_STRING': '37f332f68db77bd9d7edd4969571ad671cf9dd3b',
      'EMPTY_STRING': '9c1185a5c5e9fc54612808977ee8f548b2258d31',
      }),
    ('blake2b', False,
     {'TEST_STRING': 'a8add4bdddfd93e4877d2746e62817b1'
                     '16364a1fa7bc148d95090bc7333b3673'
                     'f82401cf7aa2e4cb1ecd90296e3f14cb'
                     '5413f8ed77be73045b13914cdcd6a918',
      'EMPTY_STRING': '786a02f742015903c6c6fd852552d272'
                      '912f4740e15847618a86e217f71f5419'
                      'd25e1031afee585313896444934eb04b'
                      '903a685b1448b755d56f701afe9be2ce',
      }),
    ('blake2s', False,
     {'TEST_STRING': '606beeec743ccbeff6cbcdf5d5302aa8'
                     '55c256c29b88c8ed331ea1a6bf3c8812',
      'EMPTY_STRING': '69217a3079908094e11121d042354a7c'
                      '1f55b6482ca1a51e1b250dfd1ed0eef9',
      }),
    ('sha3_224', False,
     {'TEST_STRING': 'd15dadceaa4d5d7bb3b48f446421'
                     'd542e08ad8887305e28d58335795',
      'EMPTY_STRING': '6b4e03423667dbb73b6e15454f0e'
                      'b1abd4597f9a1b078e3f5b5a6bc7',
      }),
    ('sha3_256', False,
     {'TEST_STRING': '69070dda01975c8c120c3aada1b28239'
                     '4e7f032fa9cf32f4cb2259a0897dfc04',
      'EMPTY_STRING': 'a7ffc6f8bf1ed76651c14756a061d662'
                      'f580ff4de43b49fa82d80a4b80f8434a',
      }),
    ('sha3_384', False,
     {'TEST_STRING': '7063465e08a93bce31cd89d2e3ca8f60'
                     '2498696e253592ed26f07bf7e703cf32'
                     '8581e1471a7ba7ab119b1a9ebdf8be41',
      'EMPTY_STRING': '0c63a75b845e4f7d01107d852e4c2485'
                      'c51a50aaaa94fc61995e71bbee983a2a'
                      'c3713831264adb47fb6bd1e058d5f004',
      }),
    ('sha3_512', False,
     {'TEST_STRING': '01dedd5de4ef14642445ba5f5b97c15e'
                     '47b9ad931326e4b0727cd94cefc44fff'
                     '23f07bf543139939b49128caf436dc1b'
                     'dee54fcb24023a08d9403f9b4bf0d450',
      'EMPTY_STRING': 'a69f73cca23a9ac5c8b567dc185a756e'
                      '97c982164fe25859e0d1dcc1475c80a6'
                      '15b2123af1f5f94c11e3e9402c3ac558'
                      'f500199d95b6d3e301758586281dcd26',
      }),
    ('whirlpool', False,
     {'TEST_STRING': 'b97de512e91e3828b40d2b0fdce9ceb3'
                     'c4a71f9bea8d88e75c4fa854df36725f'
                     'd2b52eb6544edcacd6f8beddfea403cb'
                     '55ae31f03ad62a5ef54e42ee82c3fb35',
      'EMPTY_STRING': '19fa61d75522a4669b44e39c1d2e1726'
                      'c530232130d407f89afee0964997f7a7'
                      '3e83be698b288febcf88e3e03c4f0757'
                      'ea8964e59b63d93708b138cc42a66eb3',
      }),
    ('__size__', True,
     {'TEST_STRING': 43,
      'EMPTY_STRING': 0,
      }),
]


@pytest.mark.parametrize('name,required,test_hashes',
                         HASH_VARIANTS)
@pytest.mark.parametrize('test_var', ['TEST_STRING', 'EMPTY_STRING'])
def test_hash_bytes(name, required, test_hashes, test_var):
    try:
        assert hash_bytes(globals()[test_var], name) == test_hashes[test_var]
    except UnsupportedHash:
        if not required:
            pytest.skip(f'Hash {name} not supported')
        raise


@pytest.mark.parametrize('test_var', ['TEST_STRING', 'EMPTY_STRING'])
def test_hash_file(test_var):
    with io.BytesIO(globals()[test_var]) as f:
        assert (hash_file(f, REQUIRED_TEST_HASHES[test_var].keys()) ==
                REQUIRED_TEST_HASHES[test_var])


@pytest.mark.parametrize('test_var', ['TEST_STRING', 'EMPTY_STRING'])
def test_hash_path(tmp_path, test_var):
    with open(tmp_path / 'test.txt', 'wb') as f:
        f.write(globals()[test_var])
    assert (hash_path(tmp_path / 'test.txt',
                      REQUIRED_TEST_HASHES[test_var].keys()) ==
            REQUIRED_TEST_HASHES[test_var])
